#
# Copyright (C) 2018-2025  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Admin REST interface endpoints."""

from typing import Optional
from uuid import UUID

from django.db import IntegrityError, transaction
from django.db.models import QuerySet
from django.http import HttpRequest
from django.utils import timezone
from grill import Properties
from ninja import FilterSchema, Query, Router
from ninja.pagination import paginate

from django_pain.api.order import ListOrderByBase
from django_pain.api.paginate import LimitOffsetPagination
from django_pain.api.schema import (
    AdminAssignPaymentRequest,
    AdminBankAccountCreateRequest,
    AdminBankAccountResponse,
    AdminBankAccountUpdateRequest,
    AdminBankPaymentResponse,
    AdminCustomerResponse,
    AdminInvoiceNumberCreateRequest,
    AdminInvoiceNumberResponse,
    AdminInvoiceResponse,
    AdminItemTypeCreateRequest,
    AdminItemTypeResponse,
    AdminPaymentImportHistoryResponse,
    AdminPaymentRequestResponse,
    AdminVariableSymbolCreateRequest,
    AdminVariableSymbolResponse,
    AdminVatRateResponse,
    ApplicationSchema,
    ErrorResponse,
    ListBankAccountFilter,
    ListBankAccountOrder,
    ListBankPaymentFilter,
    ListBankPaymentOrder,
    ListCustomerFilter,
    ListCustomerOrder,
    ListInvoiceFilter,
    ListInvoiceNumberFilter,
    ListInvoiceNumberOrder,
    ListInvoiceOrder,
    ListItemTypeFilter,
    ListItemTypeOrder,
    ListPaymentImportHistoryFilter,
    ListPaymentImportHistoryOrder,
    ListPaymentRequestFilter,
    ListPaymentRequestOrder,
    ListVariableSymbolFilter,
    ListVariableSymbolOrder,
    ListVatRateFilter,
    ListVatRateOrder,
    PaymentProcessorAssignHintSchema,
    PaymentProcessorSchema,
)
from django_pain.api.token import JweAuth
from django_pain.constants import (
    ApiPermissionScope,
    LogEntryType,
    LogReferenceType,
    LogResult,
    PaymentState,
    ProcessPaymentResultType,
)
from django_pain.logger import create_log_entry, make_log_list_properties
from django_pain.models import BankAccount, BankPayment, Customer, Invoice, PaymentImportHistory, PaymentRequest
from django_pain.models.invoices import BillingItemType, InvoiceNumberSequence, VariableSymbolSequence, VatRate
from django_pain.processors import InvalidTaxDateError
from django_pain.settings import SETTINGS, get_processor_instance

router = Router()


def _list_items(
    request: HttpRequest,
    queryset: QuerySet,
    filters: FilterSchema,
    order_by: ListOrderByBase,
    page: LimitOffsetPagination.Input,
    log_entry_type: LogEntryType,
) -> QuerySet:
    """Implement common functionality of admin list endpoints."""
    properties = make_log_list_properties(filters, order_by, page)
    with create_log_entry(request, log_entry_type, properties=properties) as log_entry:
        queryset = filters.filter(queryset)
        queryset = queryset.order_by(*order_by.get_arguments())
        log_entry.result = LogResult.SUCCESS
        return queryset


# --- bank account ---
@router.get(
    "/bank_account/{uid}/",
    tags=["Bank account"],
    summary="Bank account information",
    response={200: AdminBankAccountResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW),
)
def get_bank_account_info(request: HttpRequest, uid: str):
    """Get details of bank account."""
    with create_log_entry(request, LogEntryType.ACCOUNT_DETAIL, properties={"account_id": uid}) as log_entry:
        try:
            bank_account = BankAccount.objects.filter(uuid=UUID(uid)).first()
        except ValueError:
            bank_account = None
        if bank_account is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Bank account '{uid}' not found"}
        log_entry.references = {LogReferenceType.ACCOUNT: str(bank_account.uuid)}
        log_entry.result = LogResult.SUCCESS
        return 200, bank_account


@router.post(
    "/bank_account/",
    tags=["Bank account"],
    summary="Create bank account",
    response={201: AdminBankAccountResponse, 400: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_ACCOUNT),
)
def create_bank_account(request: HttpRequest, data: AdminBankAccountCreateRequest):
    """Create new bank account."""
    data_dump = data.model_dump(exclude_defaults=True)
    with create_log_entry(request, LogEntryType.ACCOUNT_CREATE, properties=data_dump) as log_entry:
        try:
            bank_account = BankAccount.objects.create(**data.model_dump(by_alias=True))
        except IntegrityError:
            log_entry.result = LogResult.FAIL
            return 400, {"details": f"Bank account '{data.number}' already exists"}
        log_entry.references = {LogReferenceType.ACCOUNT: str(bank_account.uuid)}
        log_entry.result = LogResult.SUCCESS
        return 201, bank_account


@router.delete(
    "/bank_account/{uid}/",
    tags=["Bank account"],
    summary="Delete bank account",
    response={200: None, 400: ErrorResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_ACCOUNT),
)
@transaction.atomic
def delete_bank_account(request: HttpRequest, uid: str):
    """Delete existing bank account."""
    with create_log_entry(request, LogEntryType.ACCOUNT_DELETE, properties={"account_id": uid}) as log_entry:
        try:
            bank_account = BankAccount.objects.filter(uuid=UUID(uid)).select_for_update().first()
        except ValueError:
            bank_account = None
        if bank_account is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Bank account '{uid}' not found"}
        try:
            bank_account.delete()
        except IntegrityError:
            log_entry.result = LogResult.FAIL
            return 400, {"details": f"Bank account '{uid}' not deleted because it is in use"}
        log_entry.references = {LogReferenceType.ACCOUNT: str(bank_account.uuid)}
        log_entry.result = LogResult.SUCCESS
        return 200, None


@router.put(
    "/bank_account/{uid}/",
    tags=["Bank account"],
    summary="Update bank account",
    response={200: AdminBankAccountResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_ACCOUNT),
)
@transaction.atomic
def update_bank_account(request: HttpRequest, uid: str, data: AdminBankAccountUpdateRequest):
    """Update existing bank account."""
    properties = {"account_id": uid}
    properties.update(data.model_dump(mode="json"))
    with create_log_entry(request, LogEntryType.ACCOUNT_UPDATE, properties=properties) as log_entry:
        try:
            bank_account = BankAccount.objects.filter(uuid=UUID(uid)).select_for_update().first()
        except ValueError:
            bank_account = None
        if bank_account is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Bank account '{uid}' not found"}
        bank_account.account_name = data.name
        bank_account.full_clean()
        bank_account.save(update_fields=["account_name"])
        log_entry.references = {LogReferenceType.ACCOUNT: str(bank_account.uuid)}
        log_entry.result = LogResult.SUCCESS
        return 200, bank_account


@router.get(
    "bank_accounts/",
    tags=["Bank account"],
    summary="List bank accounts",
    response=list[AdminBankAccountResponse],
    auth=JweAuth(ApiPermissionScope.VIEW),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_bank_acccounts(
    request: HttpRequest,
    filters: ListBankAccountFilter = Query(...),
    order_by: ListBankAccountOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List bank accounts with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=BankAccount.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.ACCOUNT_LIST,
    )


# --- bank payment ---
@router.get(
    "/bank_payment/{uid}/",
    tags=["Bank payment"],
    summary="Bank payment information",
    response={200: AdminBankPaymentResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW),
)
def get_bank_payment_info(request: HttpRequest, uid: str):
    """Get details of bank payment."""
    with create_log_entry(request, LogEntryType.PAYMENT_DETAIL, properties={"payment_id": uid}) as log_entry:
        try:
            bank_payment = BankPayment.objects.filter(uuid=UUID(uid)).select_related("account", "client_ref").first()
        except ValueError:
            bank_payment = None
        if bank_payment is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Bank payment '{uid}' not found"}
        log_entry.references = {LogReferenceType.PAYMENT: str(bank_payment.uuid)}
        log_entry.result = LogResult.SUCCESS
        return 200, bank_payment


@router.post(
    "/bank_payment/{uid}/refund/",
    tags=["Bank payment"],
    summary="Mark bank payment refunded",
    response={200: AdminBankPaymentResponse, 400: ErrorResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_PAYMENT),
)
@transaction.atomic
def mark_bank_payment_refunded(request: HttpRequest, uid: str):
    """Change state of bank payment to refunded."""
    with create_log_entry(request, LogEntryType.PAYMENT_REFUNDED, properties={"payment_id": uid}) as log_entry:
        try:
            bank_payment = (
                BankPayment.objects.filter(uuid=UUID(uid)).select_related("account").select_for_update().first()
            )
        except ValueError:
            bank_payment = None
        if bank_payment is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Bank payment '{uid}' not found"}
        if bank_payment.state not in (PaymentState.TO_REFUND, PaymentState.DEFERRED):
            log_entry.result = LogResult.FAIL
            return 400, {"details": f"Bank payment '{uid}' not ready for refunding"}
        bank_payment.state = PaymentState.REFUNDED
        bank_payment.full_clean()
        bank_payment.save(update_fields=["state"])
        log_entry.references = {LogReferenceType.PAYMENT: str(bank_payment.uuid)}
        log_entry.result = LogResult.SUCCESS

        return 200, bank_payment


@router.post(
    "/bank_payment/{uid}/assign/",
    tags=["Bank payment"],
    summary="Assign payment for processing",
    response={200: AdminBankPaymentResponse, 400: ErrorResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_PAYMENT),
)
@transaction.atomic
def assign_bank_payment(request: HttpRequest, uid: str, params: AdminAssignPaymentRequest):
    """Manually assign payment to handle with provided processor."""
    properties: Properties = {"payment_id": uid}
    properties.update(params.model_dump(mode="json", exclude_defaults=True))
    with create_log_entry(request, LogEntryType.PAYMENT_ASSIGN, properties=properties) as log_entry:
        try:
            bank_payment = (
                BankPayment.objects.filter(uuid=UUID(uid)).select_related("account").select_for_update().first()
            )
        except ValueError:
            bank_payment = None
        if bank_payment is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Bank payment '{uid}' not found"}
        if bank_payment.state not in (PaymentState.READY_TO_PROCESS, PaymentState.DEFERRED):
            log_entry.result = LogResult.FAIL
            return 400, {"details": f"Bank payment in state '{bank_payment.state}' - not bound for processing"}

        processor = get_processor_instance(params.processor)
        kwargs = {"tax_date": params.tax_date} if processor.manual_tax_date else {}
        try:
            result = processor.assign_payment(bank_payment, params.hint, **kwargs)
        except InvalidTaxDateError:
            log_entry.result = LogResult.FAIL
            return 400, {"details": "Invalid tax date for assignment"}
        if result.result == ProcessPaymentResultType.PROCESSED:
            bank_payment.processor = params.processor
            bank_payment.state = PaymentState.PROCESSED
        elif result.result == ProcessPaymentResultType.TO_REFUND:
            bank_payment.processor = params.processor
            bank_payment.state = PaymentState.TO_REFUND
        else:
            assert result.unprocessed
            log_entry.result = LogResult.FAIL
            return 400, {"details": f"Payment not processed by '{params.processor}' processor"}
        bank_payment.full_clean()
        bank_payment.save()
        log_entry.references = {LogReferenceType.PAYMENT: str(bank_payment.uuid)}
        log_entry.result = LogResult.SUCCESS

        return 200, bank_payment


@router.get(
    "bank_payments/",
    tags=["Bank payment"],
    summary="List bank payments",
    response=list[AdminBankPaymentResponse],
    auth=JweAuth(ApiPermissionScope.VIEW),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_bank_payments(
    request: HttpRequest,
    filters: ListBankPaymentFilter = Query(...),
    order_by: ListBankPaymentOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List bank payments with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=BankPayment.objects.select_related("account", "client_ref").all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.PAYMENT_LIST,
    )


# --- payment import history ---
@router.get(
    "/payment_import_history/{uid}/",
    tags=["Payment import history"],
    summary="Payment import history information",
    response={200: AdminPaymentImportHistoryResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW_IMPORT),
)
def get_payment_import_history_info(request: HttpRequest, uid: str):
    """Get details of payment import history item."""
    with create_log_entry(request, LogEntryType.IMPORT_DETAIL, properties={"import_id": uid}) as log_entry:
        try:
            payment_import_history = PaymentImportHistory.objects.filter(uuid=UUID(uid)).first()
        except ValueError:
            payment_import_history = None
        if payment_import_history is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Payment import history item '{uid}' not found"}
        log_entry.result = LogResult.SUCCESS
        return 200, payment_import_history


@router.get(
    "/payment_import_history/",
    tags=["Payment import history"],
    summary="Payment import history list",
    response=list[AdminPaymentImportHistoryResponse],
    auth=JweAuth(ApiPermissionScope.VIEW_IMPORT),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_payment_import_history(
    request: HttpRequest,
    filters: ListPaymentImportHistoryFilter = Query(...),
    order_by: ListPaymentImportHistoryOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List payment import history with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=PaymentImportHistory.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.IMPORT_LIST,
    )


# --- customer ---
@router.get(
    "/customer/{uid}/",
    tags=["Customer"],
    summary="Customer information",
    response={200: AdminCustomerResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW),
)
def get_customer_info(request: HttpRequest, uid: str):
    """Get details of customer."""
    with create_log_entry(request, LogEntryType.CUSTOMER_DETAIL, properties={"customer_id": uid}) as log_entry:
        customer = Customer.objects.filter(uid=uid).first()
        if customer is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Customer '{uid}' not found"}
        log_entry.references = {LogReferenceType.CUSTOMER: str(customer.uid)}
        log_entry.result = LogResult.SUCCESS
        return 200, customer


@router.get(
    "customers/",
    tags=["Customer"],
    summary="List customers",
    response=list[AdminCustomerResponse],
    auth=JweAuth(ApiPermissionScope.VIEW),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_customers(
    request: HttpRequest,
    filters: ListCustomerFilter = Query(...),
    order_by: ListCustomerOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List customers with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=Customer.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.CUSTOMER_LIST,
    )


# --- payment request ---
@router.get(
    "/payment_request/{uid}/",
    tags=["Payment request"],
    summary="Payment request information",
    response={200: AdminPaymentRequestResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW),
)
def get_payment_request_info(request: HttpRequest, uid: str):
    """Get details of payment request."""
    with create_log_entry(request, LogEntryType.PAYMENT_REQUEST_DETAIL, properties={"prequest_id": uid}) as log_entry:
        try:
            payment_request = PaymentRequest.objects.filter(uuid=UUID(uid)).first()
        except ValueError:
            payment_request = None
        if payment_request is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Payment request '{uid}' not found"}
        log_entry.references = {LogReferenceType.PAYMENT_REQUEST: str(payment_request.uuid)}
        log_entry.result = LogResult.SUCCESS
        return 200, payment_request


@router.post(
    "/payment_request/{uid}/close/",
    tags=["Payment request"],
    summary="Close payment request",
    response={200: AdminPaymentRequestResponse, 400: ErrorResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_PREQUEST),
)
@transaction.atomic
def close_payment_request(request: HttpRequest, uid: str):
    """Close payment request."""
    with create_log_entry(request, LogEntryType.PAYMENT_REQUEST_CLOSE, properties={"prequest_id": uid}) as log_entry:
        try:
            payment_request = PaymentRequest.objects.filter(uuid=UUID(uid)).select_for_update().first()
        except ValueError:
            payment_request = None
        if payment_request is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Payment request '{uid}' not found"}
        if payment_request.closed_at is not None:
            log_entry.result = LogResult.FAIL
            return 400, {"details": f"Payment request '{uid}' is already closed"}

        payment_request.closed_at = timezone.now()
        payment_request.full_clean()
        payment_request.save()
        log_entry.references = {LogReferenceType.PAYMENT_REQUEST: str(payment_request.uuid)}
        log_entry.result = LogResult.SUCCESS

        return 200, payment_request


@router.get(
    "payment_requests/",
    tags=["Payment request"],
    summary="List payment requests",
    response=list[AdminPaymentRequestResponse],
    auth=JweAuth(ApiPermissionScope.VIEW),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_payment_requests(
    request: HttpRequest,
    filters: ListPaymentRequestFilter = Query(...),
    order_by: ListPaymentRequestOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List payment requests with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=PaymentRequest.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.PAYMENT_REQUEST_LIST,
    )


# --- invoice ---
@router.get(
    "/invoice/{uid}/",
    tags=["Invoice"],
    summary="Invoice information",
    response={200: AdminInvoiceResponse, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.VIEW),
)
def get_invoice_info(request: HttpRequest, uid: str):
    """Get details of invoice."""
    with create_log_entry(request, LogEntryType.INVOICE_DETAIL, properties={"invoice_id": uid}) as log_entry:
        try:
            invoice = Invoice.objects.filter(uuid=UUID(uid)).first()
        except ValueError:
            invoice = None
        if invoice is None:
            log_entry.result = LogResult.FAIL
            return 404, {"details": f"Invoice '{uid}' not found"}
        log_entry.references = {LogReferenceType.INVOICE: str(invoice.uuid)}
        log_entry.result = LogResult.SUCCESS
        return 200, invoice


@router.get(
    "invoices/",
    tags=["Invoice"],
    summary="List invoices",
    response=list[AdminInvoiceResponse],
    auth=JweAuth(ApiPermissionScope.VIEW),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_invoices(
    request: HttpRequest,
    filters: ListInvoiceFilter = Query(...),
    order_by: ListInvoiceOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List invoices with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=Invoice.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.INVOICE_LIST,
    )


# --- variable symbol sequence ---
@router.post(
    "/variable_symbol/",
    tags=["Variable symbol"],
    summary="Create variable symbol sequence",
    response={201: AdminVariableSymbolResponse, 400: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.MAINTAIN),
)
def create_variable_symbol(request: HttpRequest, data: AdminVariableSymbolCreateRequest):
    """Create new variable symbol sequence."""
    data_dump = data.model_dump()
    with create_log_entry(request, LogEntryType.VARIABLE_SYMBOL_CREATE, properties=data_dump) as log_entry:
        try:
            vs_sequence = VariableSymbolSequence.objects.create(**data_dump)
        except IntegrityError:
            log_entry.result = LogResult.FAIL
            return 400, {"details": "Variable symbol sequence already exists"}
        log_entry.result = LogResult.SUCCESS
        return 201, vs_sequence


@router.get(
    "variable_symbols/",
    tags=["Variable symbol"],
    summary="List variable symbol sequences",
    response=list[AdminVariableSymbolResponse],
    auth=JweAuth(ApiPermissionScope.MAINTAIN),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_variable_symbols(
    request: HttpRequest,
    filters: ListVariableSymbolFilter = Query(...),
    order_by: ListVariableSymbolOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List variable symbol sequences with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=VariableSymbolSequence.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.VARIABLE_SYMBOL_LIST,
    )


# --- invoice number sequence ---
@router.post(
    "/invoice_number/",
    tags=["Invoice number"],
    summary="Create invoice number sequence",
    response={201: AdminInvoiceNumberResponse, 400: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.MAINTAIN),
)
def create_invoice_number(request: HttpRequest, data: AdminInvoiceNumberCreateRequest):
    """Create new invoice number sequence."""
    data_dump = data.model_dump()
    with create_log_entry(request, LogEntryType.INVOICE_NUMBER_CREATE, properties=data_dump) as log_entry:
        try:
            invoice_number = InvoiceNumberSequence.objects.create(**data_dump)
        except IntegrityError:
            log_entry.result = LogResult.FAIL
            return 400, {"details": "Invoice number sequence already exists"}
        log_entry.result = LogResult.SUCCESS
        return 201, invoice_number


@router.get(
    "invoice_numbers/",
    tags=["Invoice number"],
    summary="List invoice number sequences",
    response=list[AdminInvoiceNumberResponse],
    auth=JweAuth(ApiPermissionScope.MAINTAIN),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_invoice_numbers(
    request: HttpRequest,
    filters: ListInvoiceNumberFilter = Query(...),
    order_by: ListInvoiceNumberOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List invoice number sequences with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=InvoiceNumberSequence.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.INVOICE_NUMBER_LIST,
    )


# --- billing item type ---
@router.post(
    "/item_type/",
    tags=["Item type"],
    summary="Create billing item type",
    response={201: AdminItemTypeResponse, 400: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.MAINTAIN),
)
def create_item_type(request: HttpRequest, data: AdminItemTypeCreateRequest):
    """Create new billing item type."""
    data_dump = data.model_dump()
    with create_log_entry(request, LogEntryType.ITEM_TYPE_CREATE, properties=data_dump) as log_entry:
        try:
            item_type = BillingItemType.objects.create(**data_dump)
        except IntegrityError:
            log_entry.result = LogResult.FAIL
            return 400, {"details": f"Billing item type '{data.handle}' already exists"}
        log_entry.result = LogResult.SUCCESS
        return 201, item_type


@router.get(
    "item_types/",
    tags=["Item type"],
    summary="List billing item types",
    response=list[AdminItemTypeResponse],
    auth=JweAuth(ApiPermissionScope.MAINTAIN),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_item_types(
    request: HttpRequest,
    filters: ListItemTypeFilter = Query(...),
    order_by: ListItemTypeOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List billing item types with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=BillingItemType.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.ITEM_TYPE_LIST,
    )


# --- VAT rate ---
@router.get(
    "vat_rates/",
    tags=["VAT rate"],
    summary="List VAT rates",
    response=list[AdminVatRateResponse],
    auth=JweAuth(ApiPermissionScope.MAINTAIN),
)
@paginate(LimitOffsetPagination, pass_parameter="page")
def list_vat_rates(
    request: HttpRequest,
    filters: ListVatRateFilter = Query(...),
    order_by: ListVatRateOrder = Query(...),
    page: Optional[LimitOffsetPagination.Input] = None,
):
    """List VAT rates with filtering and ordering."""
    return _list_items(
        request=request,
        queryset=VatRate.objects.all(),
        filters=filters,
        order_by=order_by,
        page=page or LimitOffsetPagination.Input(),  # type: ignore[call-arg]
        log_entry_type=LogEntryType.VAT_RATE_LIST,
    )


# --- payment processor ---
@router.get(
    "/payment_processors/",
    tags=["Payment processor"],
    summary="List of payment processors",
    response=list[PaymentProcessorSchema],
)
def list_payment_processors(request: HttpRequest):
    """List configured payment processsors."""
    processors: list[PaymentProcessorSchema] = []
    for name in SETTINGS.processors:
        processor = get_processor_instance(name)
        processors.append(
            PaymentProcessorSchema(
                name=name, objective=str(processor.default_objective), manual_tax_date=processor.manual_tax_date
            )
        )
    return processors


@router.get(
    "/payment_processor/{name}/hints/",
    tags=["Payment processor"],
    summary="Hints for payment processors",
    response={200: PaymentProcessorAssignHintSchema, 404: ErrorResponse},
    auth=JweAuth(ApiPermissionScope.EDIT_PAYMENT),
)
def payment_processor_hint(request: HttpRequest, name: str):
    """Get hints for manual assignment of payment to processor."""
    if name not in SETTINGS.processors:
        return 404, {"details": "Processor not found"}
    processor = get_processor_instance(name)
    choices = processor.get_client_choices() if hasattr(processor, "get_client_choices") else None
    return 200, PaymentProcessorAssignHintSchema(choices=choices)


# --- applications ---
@router.get(
    "/applications/",
    tags=["Application"],
    summary="List of applications for payment requests and invoices",
    response=list[ApplicationSchema],
)
def list_applications(request: HttpRequest):
    """List available applications."""
    # List of applications is hard-coded at the moment (which is going to change in future).
    applications = [
        ApplicationSchema(app_id="auctions", objective="Auctions of domain registration right"),
    ]
    return applications
