#
# Copyright (C) 2024-2025  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Support for REST API lists ordering."""

from collections.abc import Iterable

from ninja import Field, Schema
from pydantic import model_validator
from typing_extensions import Self


class ListOrderByBase(Schema):
    """Base class for ordering lists of items.

    Descendant class should provide type of response list item (to get possible ordering alias). Optionally it
    can also provide default list ordering.

    Items supposed to allow ordering should have `order_by` specified in the model field definition. It is expected
    to provide a proper parameter used in queryset ordering.
    """

    order_by: str = Field(
        default="",
        description="Comma separated fields for ordering (with '+' or '-' prefix for ascending/descending order)",
    )
    _model: type[Schema]
    _default: str = ""

    @model_validator(mode="after")
    def fields_validator(self) -> Self:
        """Check for valid order-by fields."""
        fields = self.order_by.split(",")
        for field in fields:
            if not field:
                continue
            if field[0] not in ("+", "-"):
                raise ValueError(f"Field '{field}' is not prefixed with ordering direction sign")
            field = field[1:]
            if field not in self._model.model_fields:  # type: ignore[operator]
                raise ValueError(f"Field '{field}' is not included in model '{self._model.__name__}'")
            extra = self._model.model_fields[field].json_schema_extra  # type: ignore[index]
            if not isinstance(extra, dict) or not extra.get("order_by", None):
                raise ValueError(f"Field '{field}' is not supported for ordering")
        return self

    def get_arguments(self) -> Iterable[str]:
        """Get ordering arguments for queryset."""
        order_by = self.order_by or self._default
        fields = order_by.split(",")
        for field in fields:
            if not field:
                continue
            order_by_field = self._model.model_fields[field[1:]].json_schema_extra["order_by"]  # type: ignore[index]
            assert isinstance(order_by_field, str)
            yield order_by_field if field[0] == "+" else f"-{order_by_field}"

        yield "id"  # finally order by universal primary key for ORM models
