#
# Copyright (C) 2018-2025  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Schemas for REST admin input/output values."""

from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Any, Literal, Optional, Union
from uuid import UUID

from django.db.models import Exists, OuterRef, Q
from django.utils import timezone
from django_countries import countries
from moneyed.classes import Currency, CurrencyDoesNotExist, get_currency
from ninja import Field, FilterSchema, Schema
from pydantic import field_validator, model_validator
from typing_extensions import Self

from django_pain.api.order import ListOrderByBase
from django_pain.constants import ApiPermissionScope
from django_pain.models import BankPayment, Customer, Invoice, InvoiceItem, PaymentRequest, PaymentRequestItem
from django_pain.models.invoices import VatRate
from django_pain.settings import get_processor_instance
from django_pain.utils import local_date


class ErrorResponse(Schema):
    """Error response details."""

    details: Union[str, list[str], list[dict[str, Any]]]


class CountrySchema(Schema):
    """Model for country."""

    code: str = Field(description="Two-letter ISO 3166-1 code")
    name: str = Field(description="Country name")


# ------------------------------------------------------------------------------------------------------
# access token
class TokenData(Schema):
    """Data (custom payload) in access token."""

    scopes: list[ApiPermissionScope] = Field(description="Permission scopes")
    log: Optional[str] = Field(None, description="FRED gRPC logger session id")


class TokenResponse(Schema):
    """Access token response."""

    token: str = Field(description="JWE token with permissions")


# ------------------------------------------------------------------------------------------------------
# bank account
class AdminBankAccountCreateRequest(Schema):
    """Bank account model in API create request."""

    number: str = Field(
        serialization_alias="account_number",
        description="Account identification (typically in the form ``account_number/bank_code``)",
    )
    name: str = Field(default="", serialization_alias="account_name", description="Account description")
    currency: str = Field(description="Three-letter ISO 4217 currency code")

    @field_validator("currency")
    @classmethod
    def currency_validator(cls, value: str) -> str:
        """Check for valid currency code."""
        try:
            get_currency(value)
        except CurrencyDoesNotExist:
            raise ValueError(f"Currency '{value}' is not supported") from None
        return value


class AdminBankAccountUpdateRequest(Schema):
    """Bank account model in API update request."""

    # Note: Account name and currency cannot be updated as it might break payment processing workflow, etc.
    name: str = Field(serialization_alias="account_name", description="Account description")


class AdminBankAccountResponse(Schema):
    """Bank account model in API responses."""

    uid: UUID = Field(alias="uuid", description="Bank account identifier")
    number: str = Field(
        alias="account_number",
        description="Account identification (typically in the form ``account_number/bank_code``)",
    )
    name: str = Field(
        alias="account_name", description="Account description", json_schema_extra={"order_by": "account_name"}
    )
    currency: str = Field(description="Three-letter ISO 4217 currency code", json_schema_extra={"order_by": "currency"})


class ListBankAccountFilter(FilterSchema):
    """Filters in bank account list."""

    number: Optional[str] = Field(
        None, description="Filter by substring in account number", json_schema_extra={"q": "account_number__icontains"}
    )
    name: Optional[str] = Field(
        None, description="Filter by substring in account name", json_schema_extra={"q": "account_name__icontains"}
    )
    currency: Optional[str] = Field(
        None,
        description="Filter by currency (case insensitive full match)",
        json_schema_extra={"q": "currency__iexact"},
    )


class ListBankAccountOrder(ListOrderByBase):
    """Ordering in bank account list.

    Fields with support for ordering: ``name``, ``currency``.
    """

    _model = AdminBankAccountResponse


# ------------------------------------------------------------------------------------------------------
# bank payment
class ClientRefSchema(Schema):
    """Client model to refer to from bank payment."""

    handle: str = Field(description="Client handle")


class InvoiceRefSchema(Schema):
    """Invoice model to refer to from bank payment."""

    number: str = Field(description="Invoice number")
    file_uid: Optional[str] = Field(description="File identifier in remote system")
    invoice_type: str = Field(description="Type of invoice")


class AdminBankPaymentResponse(Schema):
    """Bank payment model in API responses."""

    # identifiers
    uuid: UUID = Field(description="Unique identifier")
    identifier: str = Field(description="Identifier assigned by bank", json_schema_extra={"order_by": "identifier"})
    # payment data
    account_number: str = Field(
        alias="account.account_number",
        description="Bank account number (our)",
        json_schema_extra={"order_by": "account__account_number"},
    )
    counter_account_number: str = Field(
        description="Counter account number", json_schema_extra={"order_by": "counter_account_number"}
    )
    counter_account_name: str = Field(description="Counter account name")
    amount: Decimal = Field(
        alias="amount.amount", description="Payment amount", json_schema_extra={"order_by": "amount"}
    )
    currency: str = Field(
        alias="amount.currency",
        description="Payment currency (three-letter ISO 4217 code)",
        json_schema_extra={"order_by": "amount_currency"},
    )
    transaction_date: Optional[date] = Field(
        description="Date of transaction", json_schema_extra={"order_by": "transaction_date"}
    )
    constant_symbol: str = Field(description="Constant symbol of payment")
    variable_symbol: str = Field(
        description="Variable symbol of payment",
        json_schema_extra={"order_by": "raw_variable_symbol"},
    )
    specific_symbol: str = Field(description="Specific symbol of payment")
    description: str = Field(description="Payment memo provided by payer or bank")
    # processing
    created: datetime = Field(
        alias="create_time",
        description="Time point of payment creation in DB",
        json_schema_extra={"order_by": "create_time"},
    )
    payment_type: str = Field(description="Payment type - bank transfer or card payment")
    state: str = Field(
        description="State of payment processing - initialized, ready to process, processed, ...",
        json_schema_extra={"order_by": "state"},
    )
    processor: str = Field(description="Processor responsible for handling payment")
    processing_error: Optional[str] = Field(description="Error encounter during processing payment")
    # card payment related
    card_payment_state: str = Field(description="State of card payment declared by bank gateway")
    card_handler: str = Field(description="Handler assigned to payment")
    custom_expiry: Optional[datetime] = Field(description="Expiry of custom card payment")
    customer_code: str = Field(description="Received code for custom card payment")
    email: Optional[str] = Field(description="Customer email (for custom card payments)")
    email_sent: Optional[datetime] = Field(description="Time point of email notification with customer code")
    # references
    client: Optional[ClientRefSchema]
    invoices: list[InvoiceRefSchema]

    @field_validator("currency", mode="before")
    @classmethod
    def validate_currency(cls, value: Any) -> Any:
        """Validate currency field."""
        return str(value) if isinstance(value, Currency) else value

    @staticmethod
    def resolve_variable_symbol(payment: BankPayment) -> str:
        """Return normalized variable symbol."""
        return payment.variable_symbol

    @staticmethod
    def resolve_client(payment: BankPayment) -> Optional[ClientRefSchema]:
        """Return related client (if available)."""
        try:
            client_ref = payment.client_ref
            return ClientRefSchema(handle=client_ref.handle)
        except BankPayment.client_ref.RelatedObjectDoesNotExist:
            return None

    @staticmethod
    def resolve_invoices(payment: BankPayment) -> list[InvoiceRefSchema]:
        """Return related invoices (if available)."""
        return [
            InvoiceRefSchema(
                number=item.number,
                file_uid=None,  # ToDo later
                invoice_type=item.invoice_type,
            )
            for item in payment.invoice_refs.all()
        ]


class AdminAssignPaymentRequest(Schema):
    """Parameters for assigninment payment for processing."""

    processor: str = Field(description="Processor to handle payment")
    hint: str = Field(default="", description="Parameter for payment processing (depends on processor)")
    tax_date: Optional[date] = Field(default=None, description="Tax date (applicable to some processors only)")

    @model_validator(mode="after")
    def check_processor(self) -> Self:
        """Check processor name, compatibility of other parameters."""
        processor = get_processor_instance(self.processor)  # raises ValueError on problem
        if processor.manual_tax_date and self.tax_date is None:
            raise ValueError(f"Tax date for processor '{self.processor}' is required")
        return self


class ListBankPaymentFilter(FilterSchema):
    """Filters in bank payment list."""

    identifier: Optional[str] = Field(None, description="Filter by bank identifier")
    account_number: Optional[str] = Field(
        None,
        description="Filter by (own) account number (exact match)",
        json_schema_extra={"q": "account__account_number__exact"},
    )
    counter_account_number: Optional[str] = Field(None, description="Filter by counter account number (exact match)")
    counter_account_name: Optional[str] = Field(
        None,
        description="Filter by counter account number (case insensitive substring)",
        json_schema_extra={"q": "counter_account_name__icontains"},
    )
    amount: Optional[Decimal] = Field(None, description="Filter by payment amount (numerical equality)")
    currency: Optional[str] = Field(
        None,
        description="Filter by payment currency (case insensitive full match)",
        json_schema_extra={"q": "amount_currency__iexact"},
    )
    transaction_date_from: Optional[date] = Field(
        None, description="Filter by transaction date (lower bound)", json_schema_extra={"q": "transaction_date__gte"}
    )
    transaction_date_to: Optional[date] = Field(
        None, description="Filter by transaction date (upper bound)", json_schema_extra={"q": "transaction_date__lte"}
    )
    variable_symbol: Optional[str] = Field(None, description="Filter by variable symbol (significant digits match)")
    state: Optional[list[str]] = Field(
        None, description="Filter by payment states (repeated values)", json_schema_extra={"q": "state__in"}
    )
    description: Optional[str] = Field(
        None,
        description="Filter by payment memo (case insensitive substring)",
        json_schema_extra={"q": "description__icontains"},
    )

    def filter_variable_symbol(self, value: Optional[str]) -> Q:
        """Filter by variable symbol ragardless of leading zeroes."""
        return Q() if value is None else Q(raw_variable_symbol__regex=r"^0*" + value.lstrip("0") + r"$")


class ListBankPaymentOrder(ListOrderByBase):
    """Ordering in bank payment list.

    Fields with support for ordering: ``identifier``, ``account_number``, ``counter_account_number``, ``amount``,
    ``currency``, ``transacton_date``, ``variable_symbol``, ``state``.
    """

    _model = AdminBankPaymentResponse
    _default = "-created"  # most recent first


# ------------------------------------------------------------------------------------------------------
# payment import history
class AdminPaymentImportHistoryResponse(Schema):
    """Payment import history model in API responses."""

    uuid: UUID = Field(description="Unique identifier")
    origin: str = Field(description="Source of import", json_schema_extra={"order_by": "origin"})
    start_datetime: datetime = Field(description="Start of import", json_schema_extra={"order_by": "start_datetime"})
    filenames: list[str] = Field(description="List of imported filenames (not applicable for all sources)")
    errors: Optional[int] = Field(description="Number of errors during import")
    finished: bool = Field(description="Flag to indicate that import was completed")


class ListPaymentImportHistoryFilter(FilterSchema):
    """Filters in payment import history list."""

    origin: Optional[str] = Field(None, description="Filter by import origin")
    start_datetime_from: Optional[datetime] = Field(
        None, description="Filter by start datetime (lower bound)", json_schema_extra={"q": "start_datetime__gte"}
    )
    start_datetime_to: Optional[datetime] = Field(
        None, description="Filter by start datetime (upper bound)", json_schema_extra={"q": "start_datetime__lt"}
    )
    filenames: Optional[str] = Field(
        None,
        description="Filter by filenames (case insensitive substring)",
        json_schema_extra={"q": "_filenames__icontains"},
    )
    finished: Optional[bool] = Field(None, description="Filter by finish status")


class ListPaymentImportHistoryOrder(ListOrderByBase):
    """Ordering in bank payment list.

    Fields with support for ordering: ``origin``, ``start_datetime``.
    """

    _model = AdminPaymentImportHistoryResponse
    _default = "-start_datetime"  # most recent first


# ------------------------------------------------------------------------------------------------------
# customer
class AdminCustomerResponse(Schema):
    """Customer model in API responses."""

    uid: str = Field(description="Customer identifier")
    name: Optional[str] = Field(
        description="Name (physical persons or organization representative)", json_schema_extra={"order_by": "name"}
    )
    organization: Optional[str] = Field(description="Organization name", json_schema_extra={"order_by": "organization"})
    street: list[str]
    city: str = Field(description="City")
    state_or_province: Optional[str] = Field(description="State or province")
    postal_code: str = Field(description="Postal code")
    country: CountrySchema = Field(description="Customer country", json_schema_extra={"order_by": "country_code"})
    email: Optional[str] = Field(description="Email address")
    phone_number: Optional[str] = Field(description="Phone number")
    company_registration_number: Optional[str] = Field(
        description="Company registration number", json_schema_extra={"order_by": "company_registration_number"}
    )
    vat_number: Optional[str] = Field(description="VAT number", json_schema_extra={"order_by": "vat_number"})
    vies_vat_valid: Optional[bool] = Field(
        description="Result of VAT number verificatin with VIES service",
        json_schema_extra={"order_by": "vies_vat_valid"},
    )
    vies_checked_at: Optional[datetime] = Field(
        description="Datetime of last VAT number verifiction with VIES service",
        json_schema_extra={"order_by": "vies_checked_at"},
    )

    @staticmethod
    def resolve_street(item: Customer) -> list[str]:
        """Return customer street as list of parts."""
        return [i for i in (item.street1, item.street2, item.street3) if i]

    @staticmethod
    def resolve_country(item: Customer) -> CountrySchema:
        """Return customer country."""
        code = str(item.country_code)
        return CountrySchema(code=code, name=countries.name(code))


class ListAddressFilter(FilterSchema):
    """Filters in address part of customer, payment request and invoice."""

    name: Optional[str] = Field(
        None,
        description="Filter by name or organization (case insensitive substring)",
        json_schema_extra={"q": ["name__icontains", "organization__icontains"]},
    )
    address: Optional[str] = Field(
        None,
        description="Filter by address - street, city, state/province (case insensitive substring)",
        json_schema_extra={
            "q": [
                "street1__icontains",
                "street2__icontains",
                "street3__icontains",
                "city__icontains",
                "state_or_province__icontains",
            ]
        },
    )
    postal_code: Optional[str] = Field(None, description="Filter by postal code (full match)")
    country: Optional[str] = Field(
        None,
        description="Filter by country code (case insensitive full match)",
        json_schema_extra={"q": "country_code__iexact"},
    )
    company_registration_number: Optional[str] = Field(
        None,
        description="Filter by company registration number (case insensitive full match)",
        json_schema_extra={"q": "company_registration_number__iexact"},
    )
    vat_number: Optional[str] = Field(
        None,
        description="Filter by VAT number (case insensitive full match)",
        json_schema_extra={"q": "vat_number__iexact"},
    )


class ListCustomerFilter(ListAddressFilter):
    """Filters in customer list."""

    email: Optional[str] = Field(
        None,
        description="Filter by email address (case insensitive full match)",
        json_schema_extra={"q": "email__iexact"},
    )
    vies_vat_valid: Optional[bool] = Field(None, description="Filter by VIES check status")


class ListCustomerOrder(ListOrderByBase):
    """Ordering in customer list.

    Fields with support for ordering: ``name``, ``organization``, ``country``, ``company_registration_number``,
    ``vat_number``, ``vies_vat_valid``, ``vies_checked_at``.
    """

    _model = AdminCustomerResponse


# ------------------------------------------------------------------------------------------------------
# payment request
class PaymentRequestItemSchema(Schema):
    """Payment request item model."""

    name: str = Field(description="Item description")
    item_type: Optional[str] = Field(description="Item type (to evaluate VAT)")
    quantity: int = Field(description="Quantity")
    price_per_unit: Decimal = Field(description="Unit price")
    price_total: Decimal = Field(description="Total price")


class AdminPaymentRequestResponse(Schema):
    """Payment request model in API responses."""

    # identifiers
    uid: UUID = Field(alias="uuid", description="Payment request identifier")
    customer_uid: str = Field(alias="customer.uid", description="Customer identifier")
    app_id: str = Field(description="Application identifier", json_schema_extra={"order_by": "app_id"})
    # copy of customer info at the moment of creation
    name: Optional[str] = Field(description="Customer name", json_schema_extra={"order_by": "name"})
    organization: Optional[str] = Field(description="Organization name", json_schema_extra={"order_by": "organization"})
    street: list[str]
    city: str = Field(description="City")
    state_or_province: Optional[str] = Field(description="State or province")
    postal_code: str = Field(description="Postal code")
    country: CountrySchema = Field(description="Country", json_schema_extra={"order_by": "country_code"})
    company_registration_number: Optional[str] = Field(
        description="Company registration number", json_schema_extra={"order_by": "company_registration_number"}
    )
    vat_number: Optional[str] = Field(description="VAT number", json_schema_extra={"order_by": "vat_number"})
    # request lifetime
    created_at: datetime = Field(description="Creation time", json_schema_extra={"order_by": "created_at"})
    due_at: Optional[datetime] = Field(
        description="Due time real (may be unlimited)", json_schema_extra={"order_by": "due_at"}
    )
    due_at_displayed: datetime = Field(
        description="Due time displayed to customer", json_schema_extra={"order_by": "due_at_displayed"}
    )
    closed_at: Optional[datetime] = Field(
        description="Close time (by realized payment or manually)", json_schema_extra={"order_by": "closed_at"}
    )
    # total price to pay
    price_to_pay: Decimal = Field(description="Total price to pay", json_schema_extra={"order_by": "price_to_pay"})
    currency: str = Field(description="Three-letter ISO 4217 currency code", json_schema_extra={"order_by": "currency"})
    # payment
    variable_symbol: str = Field(description="Variable symbol", json_schema_extra={"order_by": "variable_symbol"})
    account: Optional[str] = Field(
        alias="bank_account.account_number",
        description="Bank account for transfer payment",
        json_schema_extra={"order_by": "bank_account__account_number"},
    )
    card_payment: Optional[UUID] = Field(
        default=None, alias="card_payment.uuid", description="Identifier of card payment"
    )
    realized_payment: Optional[UUID] = Field(
        default=None, alias="realized_payment.uuid", description="Identifier of realized payment"
    )
    # documents
    file_uid: Optional[str] = Field(description="PDF file identifier in document storage")
    invoice_uid: Optional[UUID] = Field(default=None, alias="invoice.uuid", description="Related invoice identifier")
    # items
    items: list[PaymentRequestItemSchema] = Field(description="Payment request items")

    @staticmethod
    def resolve_street(item: PaymentRequest) -> list[str]:
        """Return customer street as list of parts."""
        return [i for i in (item.street1, item.street2, item.street3) if i]

    @staticmethod
    def resolve_country(item: PaymentRequest) -> CountrySchema:
        """Return customer country."""
        code = str(item.country_code)
        return CountrySchema(code=code, name=countries.name(code))


class ListPaymentRequestFilter(ListAddressFilter):
    """Filters in payment request list."""

    app_id: Optional[str] = Field(None, description="Filter by application identifier (full match)")
    customer_uid: Optional[str] = Field(
        None, description="Filter by customer identifier (full match)", json_schema_extra={"q": "customer__uid"}
    )
    email: Optional[str] = Field(
        None,
        description="Filter by email address (case insensitive full match)",
        json_schema_extra={"q": "customer__email__iexact"},
    )
    created_at_from: Optional[datetime] = Field(
        None, description="Filter by creation time (lower bound)", json_schema_extra={"q": "created_at__gte"}
    )
    created_at_to: Optional[datetime] = Field(
        None, description="Filter by creation time (upper bound)", json_schema_extra={"q": "created_at__lt"}
    )
    closed_at_from: Optional[datetime] = Field(
        None, description="Filter by closing time (lower bound)", json_schema_extra={"q": "closed_at__gte"}
    )
    closed_at_to: Optional[datetime] = Field(
        None, description="Filter by closing time (upper bound)", json_schema_extra={"q": "closed_at__lt"}
    )
    price_to_pay: Optional[Decimal] = Field(None, description="Filter by price to pay (numerical equality)")
    currency: Optional[str] = Field(
        None,
        description="Filter by payment currency (case insensitive full match)",
        json_schema_extra={"q": "currency__iexact"},
    )
    variable_symbol: Optional[str] = Field(None, description="Filter by variable symbol (full match)")
    state: Optional[Literal["waiting", "overdue", "realized", "closed"]] = Field(None, description="Filter by state")
    item_name: Optional[str] = Field(None, description="Filter by item name (case insensitive substring)")

    def filter_state(self, value: Optional[str]) -> Q:
        """Filter by payment request state."""
        if value == "waiting":
            return Q(realized_payment__isnull=True, closed_at__isnull=True, due_at_displayed__gte=timezone.now())
        elif value == "overdue":
            return Q(realized_payment__isnull=True, closed_at__isnull=True, due_at_displayed__lt=timezone.now())
        elif value == "realized":
            return Q(realized_payment__isnull=False)
        elif value == "closed":
            return Q(realized_payment__isnull=True, closed_at__isnull=False)
        else:
            assert value is None
            return Q()

    def filter_item_name(self, value: Optional[str]) -> Q:
        """Filter by substring existence in some item name."""
        if value is None:
            return Q()
        items = PaymentRequestItem.objects.filter(payment_request_id=OuterRef("pk"), name__icontains=value)
        return Q(Exists(items))


class ListPaymentRequestOrder(ListOrderByBase):
    """Ordering in payment request list.

    Fields with support for ordering: ``app_id``, ``name``, ``organization``, ``country``,
    ``company_registration_number``, ``vat_number``, ``created_at``, ``due_at``, ``due_at_displayed``, ``closed_at``,
    ``price_to_pay``, ``currency``, ``variable_symbol``, ``account``.
    """

    _model = AdminPaymentRequestResponse
    _default = "-created_at"  # most recent first


# ------------------------------------------------------------------------------------------------------
# invoice
class InvoiceItemSchema(Schema):
    """Invoice item model."""

    name: str = Field(description="Item description")
    item_type: Optional[str] = Field(description="Item type (to evaluate VAT)")
    vat_rate: Decimal = Field(description="VAT rate for item (in %)")
    quantity: int = Field(description="Quantity")
    price_per_unit: Decimal = Field(description="Unit price (including VAT)")
    price_total_base: Decimal = Field(description="Total price (excluding VAT)")
    price_total_vat: Decimal = Field(description="Total VAT")
    price_total: Decimal = Field(description="Total price (including VAT)")


class AdminInvoiceResponse(Schema):
    """Invoice model in API responses."""

    # identifiers
    number: str = Field(description="Invoice number", json_schema_extra={"order_by": "number"})
    uid: UUID = Field(alias="uuid", description="Invoice identifier")
    customer_uid: str = Field(alias="customer.uid", description="Customer identifier")
    app_id: str = Field(description="Application identifier", json_schema_extra={"order_by": "app_id"})
    # copy of customer info at the moment of creation
    name: Optional[str] = Field(description="Customer name", json_schema_extra={"order_by": "name"})
    organization: Optional[str] = Field(description="Organization name", json_schema_extra={"order_by": "organization"})
    street: list[str]
    city: str = Field(description="City")
    state_or_province: Optional[str] = Field(description="State or province")
    postal_code: str = Field(description="Postal code")
    country: CountrySchema = Field(description="Country", json_schema_extra={"order_by": "country_code"})
    company_registration_number: Optional[str] = Field(
        description="Company registration number", json_schema_extra={"order_by": "company_registration_number"}
    )
    vat_number: Optional[str] = Field(description="VAT number", json_schema_extra={"order_by": "vat_number"})
    # invoice dates
    created_at: datetime = Field(description="Creation time", json_schema_extra={"order_by": "created_at"})
    due_at: Optional[datetime] = Field(
        description="Due time real (may be unlimited)", json_schema_extra={"order_by": "due_at"}
    )
    tax_at: date = Field(description="Tax point", json_schema_extra={"order_by": "tax_at"})
    # payment
    paid_at: Optional[date] = Field(description="Date of realized payment", json_schema_extra={"order_by": "paid_at"})
    variable_symbol: str = Field(description="Variable symbol", json_schema_extra={"order_by": "variable_symbol"})
    account: str = Field(
        alias="bank_account",
        description="Bank account for transfer payment",
        json_schema_extra={"order_by": "bank_account"},
    )
    payment_type: str = Field(description="Payment type: transfer/card payment")
    # prices
    price_total_base: Decimal = Field(description="Total price (excluding VAT)")
    price_total_vat: Decimal = Field(description="Total VAT")
    price_total: Decimal = Field(
        description="Total price (including VAT)", json_schema_extra={"order_by": "price_total"}
    )
    price_to_pay: Decimal = Field(description="Remaining amount to pay")
    currency: str = Field(description="Three-letter ISO 4217 currency code", json_schema_extra={"order_by": "currency"})
    # documents
    file_uid: Optional[str] = Field(description="PDF file identifier in document storage")
    # items
    items: list[InvoiceItemSchema] = Field(description="Invoice items")

    @staticmethod
    def resolve_street(item: Invoice) -> list[str]:
        """Return customer street as list of parts."""
        return [i for i in (item.street1, item.street2, item.street3) if i]

    @staticmethod
    def resolve_country(item: Invoice) -> CountrySchema:
        """Return customer country."""
        code = str(item.country_code)
        return CountrySchema(code=code, name=countries.name(code))


class ListInvoiceFilter(ListAddressFilter):
    """Filters in invoice list."""

    number: Optional[str] = Field(None, description="Filter by invoice number (full match)")
    app_id: Optional[str] = Field(None, description="Filter by application identifier (full match)")
    customer_uid: Optional[str] = Field(
        None, description="Filter by customer identifier (full match)", json_schema_extra={"q": "customer__uid"}
    )
    email: Optional[str] = Field(
        None,
        description="Filter by email address (case insensitive full match)",
        json_schema_extra={"q": "customer__email__iexact"},
    )
    price_total: Optional[Decimal] = Field(None, description="Filter by total price (numerical equality)")
    currency: Optional[str] = Field(
        None,
        description="Filter by payment currency (case insensitive full match)",
        json_schema_extra={"q": "currency__iexact"},
    )
    variable_symbol: Optional[str] = Field(None, description="Filter by variable symbol (full match)")
    tax_at_from: Optional[date] = Field(
        None, description="Filter by tax point (lower bound)", json_schema_extra={"q": "tax_at__gte"}
    )
    tax_at_to: Optional[date] = Field(
        None, description="Filter by tax point (upper bound)", json_schema_extra={"q": "tax_at__lte"}
    )
    paid_at_from: Optional[date] = Field(
        None, description="Filter by paid date (lower bound)", json_schema_extra={"q": "paid_at__gte"}
    )
    paid_at_to: Optional[date] = Field(
        None, description="Filter by paid date (upper bound)", json_schema_extra={"q": "paid_at__lte"}
    )
    item_name: Optional[str] = Field(None, description="Filter by item name (case insensitive substring)")

    def filter_item_name(self, value: Optional[str]) -> Q:
        """Filter by substring existence in some item name."""
        if value is None:
            return Q()
        items = InvoiceItem.objects.filter(invoice_id=OuterRef("pk"), name__icontains=value)
        return Q(Exists(items))


class ListInvoiceOrder(ListOrderByBase):
    """Ordering in invoice list.

    Fields with support for ordering: ``number``, ``app_id``, ``name``, ``organization``, ``country``,
    ``company_registration_number``, ``vat_number``, ``created_at``, ``due_at``, ``tax_at``, ``paid_at``,
    ``variable_symbol``, ``account``, ``price_total``, ``currency``.
    """

    _model = AdminInvoiceResponse
    _default = "-created_at"  # most recent first


# ------------------------------------------------------------------------------------------------------
# variable symbol sequence
class AdminVariableSymbolCreateRequest(Schema):
    """Variable symbol sequence model in API create request."""

    app_id: str = Field(description="Application identifier", min_length=1)
    year: int = Field(description="Year", ge=2000, le=2999)
    prefix: int = Field(description="Prefix", ge=0, le=9999)  # 4-digit prefix


class AdminVariableSymbolResponse(Schema):
    """Variable symbol sequence model in API responses."""

    app_id: str = Field(description="Application identifier", json_schema_extra={"order_by": "app_id"})
    year: int = Field(description="Year", json_schema_extra={"order_by": "year"})
    prefix: int = Field(description="Prefix")
    current_number: int = Field(description="Current number")


class ListVariableSymbolFilter(FilterSchema):
    """Filters in variable symbol sequence list."""

    app_id: Optional[str] = Field(None, description="Filter by application identifier (full match)")
    year: Optional[int] = Field(None, description="Filter by year (numerical equality)")


class ListVariableSymbolOrder(ListOrderByBase):
    """Ordering in variable symbol sequence list.

    Fields with support for ordering: ``app_id``, ``year``.
    """

    _model = AdminVariableSymbolResponse
    _default = "+app_id,-year"


# ------------------------------------------------------------------------------------------------------
# invoice number sequence
class AdminInvoiceNumberCreateRequest(Schema):
    """Invoice number sequence model in API create request."""

    app_id: str = Field(description="Application identifier", min_length=1)
    year: int = Field(description="Year", ge=2000, le=2999)
    prefix: str = Field(description="Prefix", pattern=r"^\d{4}$")  # 4-digit prefix (required by Helios)


class AdminInvoiceNumberResponse(Schema):
    """Invoice number sequence model in API responses."""

    app_id: str = Field(description="Application identifier", json_schema_extra={"order_by": "app_id"})
    year: int = Field(description="Year", json_schema_extra={"order_by": "year"})
    prefix: str = Field(description="Prefix")
    current_number: int = Field(description="Current number")


class ListInvoiceNumberFilter(FilterSchema):
    """Filters in invoice number sequence list."""

    app_id: Optional[str] = Field(None, description="Filter by application identifier (full match)")
    year: Optional[int] = Field(None, description="Filter by year (numerical equality)")


class ListInvoiceNumberOrder(ListOrderByBase):
    """Ordering in invoice number sequence list.

    Fields with support for ordering: ``app_id``, ``year``.
    """

    _model = AdminInvoiceNumberResponse
    _default = "+app_id,-year"


# ------------------------------------------------------------------------------------------------------
# billing item type
class AdminItemTypeCreateRequest(Schema):
    """Billing item type model in API create request."""

    handle: str = Field(
        description="Item type handle",
        pattern=r"^[a-z]+(_[a-z]+)*$",  # lowercase words separated with underscore
    )
    not_invoiceable: bool = Field(description="Not-invoiceable items prevent invoice creation")
    cpa_code: str = Field(
        description="CPA code (to define applicable VAT rate)",
        pattern=r"^(\d\d(\.\d(\d(\.\d\d?)?)?)?)?$",  # empty or 2 to 6 digit numerical code (with dot separated couples)
    )

    @model_validator(mode="after")
    def check_cpa_code_with_not_invoiceable(self) -> Self:
        """Check empty CPA code is only allowed for not-invoiceable."""
        if not self.not_invoiceable and self.cpa_code == "":
            raise ValueError("CPA code is required for invoiceable item")
        return self


class AdminItemTypeResponse(Schema):
    """Billing item type sequence model in API responses."""

    handle: str = Field(description="Item type handle", json_schema_extra={"order_by": "handle"})
    not_invoiceable: bool = Field(description="Not-invoiceable items prevent invoice creation")
    cpa_code: str = Field(description="CPA code (to define applicable VAT rate)")


class ListItemTypeFilter(FilterSchema):
    """Filters in billing item type list."""

    handle: Optional[str] = Field(
        None, description="Filter by handle (case insensitive substring)", json_schema_extra={"q": "handle__icontains"}
    )
    not_invoiceable: Optional[bool] = Field(None, description="Filter by not-invoiceable flag")


class ListItemTypeOrder(ListOrderByBase):
    """Ordering in billing item type list.

    Fields with support for ordering: ``handle``.
    """

    _model = AdminItemTypeResponse


# ------------------------------------------------------------------------------------------------------
# VAT rates
class AdminVatRateResponse(Schema):
    """VAT rate model in API responses."""

    country: CountrySchema = Field(description="Country", json_schema_extra={"order_by": "country_code"})
    cpa_code: Optional[str] = Field(
        description="CPA code (`None` for standard rate)", json_schema_extra={"order_by": "cpa_code"}
    )
    valid_from: date = Field(description="Start of validity")
    valid_to: Optional[date] = Field(description="End of validity (`None` for unlimited)")
    value: Decimal = Field(description="VAT rate (in per cent)")

    @staticmethod
    def resolve_country(item: VatRate) -> CountrySchema:
        """Return customer country."""
        code = str(item.country_code)
        return CountrySchema(code=code, name=countries.name(code))

    @staticmethod
    def resolve_valid_from(item: VatRate) -> date:
        """Return start validity as local date."""
        return local_date(item.valid_from)

    @staticmethod
    def resolve_valid_to(item: VatRate) -> Optional[date]:
        """Return end validity as local date."""
        # End date is included in validity interval
        return local_date(item.valid_to) - timedelta(days=1) if item.valid_to else None


class ListVatRateFilter(FilterSchema):
    """Filters in VAT rate list."""

    country: Optional[str] = Field(
        None,
        description="Filter by country code (case insensitive full match)",
        json_schema_extra={"q": "country_code__iexact"},
    )
    valid_on: Optional[date] = Field(None, description="Filter by validity on date")

    def filter_valid_on(self, value: Optional[date]) -> Q:
        """Filter by validity on date."""
        if value is None:
            return Q()
        return Q(valid_from__date__lte=value) & (Q(valid_to__isnull=True) | Q(valid_to__date__gt=value))


class ListVatRateOrder(ListOrderByBase):
    """Ordering in VAT rate list.

    Fields with support for ordering: ``country``, ``cpa_code``.
    """

    _model = AdminVatRateResponse
    _default = "+country"


# ------------------------------------------------------------------------------------------------------
# payment processor
class PaymentProcessorSchema(Schema):
    """Model of payment processor."""

    name: str = Field(description="Processor name")
    objective: str = Field(description="Objective of processor")
    manual_tax_date: bool = Field(description="Support for tax date in manual payment assignment")


class PaymentProcessorAssignHintSchema(Schema):
    """Model of parameters for manual assignment to processor."""

    choices: Optional[dict[str, str]] = Field(None, description="Hints for payment assignment: {value: description}")


# ------------------------------------------------------------------------------------------------------
# application
class ApplicationSchema(Schema):
    """Model of application."""

    app_id: str = Field(description="Application identifier")
    objective: str = Field(description="Objective of application")
