#
# Copyright (C) 2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Access token endpoints."""

import json
from time import time
from typing import Any, Optional, cast

from django.http import HttpRequest
from jwcrypto.jwt import JWT, JWException, JWTExpired
from ninja import Router
from ninja.errors import HttpError
from ninja.security import HttpBearer

from django_pain.api.schema import TokenData, TokenResponse
from django_pain.constants import ApiPermissionScope
from django_pain.settings import SETTINGS, get_admin_jwt_key

router = Router(tags=["Token"])


# --- authentication with encrypted JWT in header
def create_token(data: TokenData) -> str:
    """Create JWE token."""
    key = get_admin_jwt_key()
    if key is None:  # API is not configured for JWT authentication
        return ""
    issued_at = int(time())  # times as UNIX timestamps in seconds
    # token header - readable but signed
    header = {
        "alg": "dir",
        "enc": "A256GCM",
        "typ": "JWE",
    }
    # token claims - encrypted payload
    claims = data.model_dump()  # custom fields
    claims.update(  # mandatory fields
        {
            "iat": issued_at,  # issued at
            "nbf": issued_at - SETTINGS.api_authentication["admin"]["jwt_leeway"],  # not valid before
            "exp": issued_at + SETTINGS.api_authentication["admin"]["jwt_lifetime"],  # expires at
        }
    )

    token = JWT(header=header, claims=claims)
    token.make_encrypted_token(key)
    return cast(str, token.serialize())


class TokenAuthenticationError(Exception):
    """Exception during JWE token authentication."""


def authenticate_token(token: str) -> TokenData:
    """Authenticate JWE token."""
    if not token:
        raise TokenAuthenticationError("Empty token")
    jw_key = get_admin_jwt_key()
    if jw_key is None:
        raise TokenAuthenticationError("Authentication not configured")
    jwt = JWT(
        algs=["dir", "A256GCM"],
        expected_type="JWE",
        check_claims={"exp": None, "iat": None, "nbf": None},  # check mandatory field existence
    )
    jwt.leeway = SETTINGS.api_authentication["admin"]["jwt_leeway"]
    try:
        jwt.deserialize(token, key=jw_key)
        jwt.validate(jw_key)
    except JWTExpired as e:
        raise TokenAuthenticationError("Token expired") from e
    except (TypeError, ValueError, JWException) as e:
        raise TokenAuthenticationError("Invalid token") from e
    return TokenData(**json.loads(jwt.claims))


class JweNotConfigured:
    """Class to indicate that JWE authentication is not configured."""


class JweAuth(HttpBearer):
    """Authentication with encrypted JWT in header."""

    def __init__(self, *required_scopes: ApiPermissionScope) -> None:
        super().__init__()
        self.required_set = set(required_scopes)

    def __call__(self, request: HttpRequest) -> Optional[Any]:
        """Authenticate request."""
        # Authenticate when using encrypted JWT is not configured
        if SETTINGS.api_authentication is None or SETTINGS.api_authentication["admin"] is None:
            return JweNotConfigured()  # authentication passed, sentinel object returned
        # Authenticate encrypted JWT in header
        return super().__call__(request)

    def authenticate(self, request: HttpRequest, token: str) -> TokenData:
        """Authenticate with encrypted JWT."""
        try:
            token_data = authenticate_token(token)
        except TokenAuthenticationError as e:
            raise HttpError(401, f"Unauthorized: {e}") from None

        # check permission scopes
        provided_set = set(token_data.scopes)
        if not self.required_set.issubset(provided_set):
            missing = sorted(self.required_set.difference(provided_set))
            raise HttpError(403, "Access denied - missing permission scope(s): {}".format(", ".join(missing)))
        return token_data


# --- authentication with fixed token in header
class FixedTokenNotConfigured:
    """Class to indicate that fixed token authentication is not configured."""


class FixedTokenAuth(HttpBearer):
    """Simple authentication with fixed token in header."""

    def __call__(self, request: HttpRequest) -> Optional[Any]:
        """Authenticate request."""
        # Authenticate when using fixed token is not configured
        if SETTINGS.api_authentication is None or SETTINGS.api_authentication["token"] == "":
            return FixedTokenNotConfigured()  # authentication passed, sentinel object returned
        # Authenticate fixed token in header
        return super().__call__(request)

    def authenticate(self, request: HttpRequest, token: str) -> bool:
        """Authenticate with fixed token from settings."""
        expected = SETTINGS.api_authentication["token"] if SETTINGS.api_authentication else ""
        return token == expected


# --- encrypted JWT endpoint
@router.post("/", summary="Get token", response=TokenResponse, auth=FixedTokenAuth())
def get_token(request: HttpRequest, data: TokenData):
    """Get JWE access token with provided data."""
    return 200, {"token": create_token(data)}
