#
# Copyright (C) 2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""FRED gRPC logger utilities."""

from typing import Any, Optional

from django.http import HttpRequest
from grill import Logger, LogProperty, Properties, get_logger_client
from grill.clients import AbstractLoggerClient
from grill.logger import LogEntry
from ninja import FilterSchema

from django_pain.api.order import ListOrderByBase
from django_pain.api.paginate import LimitOffsetPagination
from django_pain.constants import LOGGER_SERVICE, LOGGER_SERVICE_HANDLE, LogEntryType, LogReferenceType, LogResult
from django_pain.settings import SETTINGS

LOGGER = Logger(
    get_logger_client(SETTINGS.logger_client, **SETTINGS.logger_options),
    service=LOGGER_SERVICE,
    default_result=LogResult.ERROR,
)


def register_logger(logger_client: AbstractLoggerClient) -> None:
    """Register required logger components."""
    logger_client.register_service(LOGGER_SERVICE, handle=LOGGER_SERVICE_HANDLE)
    logger_client.register_results(LOGGER_SERVICE, results=LogResult)
    logger_client.register_log_entry_types(LOGGER_SERVICE, log_entry_types=LogEntryType)
    logger_client.register_object_reference_types(reference_types=LogReferenceType)


def create_log_entry(request: HttpRequest, log_entry_type: LogEntryType, **kwargs: Any) -> LogEntry:
    """Create log entry with proper common values."""
    # Imported here to avoid `django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.` exception
    from django_pain.api.schema import TokenData

    session_id = request.auth.log if isinstance(request.auth, TokenData) else None
    return LOGGER.create(
        log_entry_type,
        source_ip=request.META.get("REMOTE_ADDR", ""),
        session_id=session_id,
        **kwargs,  # input/output properties/references, etc.
    )


def make_log_list_properties(
    filtering: FilterSchema, ordering: ListOrderByBase, pagination: LimitOffsetPagination.Input
) -> Optional[Properties]:
    """Make log properties common for list endpoints."""
    filter_dump = {key: str(value) for key, value in filtering.model_dump(exclude_none=True).items()}
    filter_property = ("filter", LogProperty("", children=filter_dump) if filter_dump else None)
    order_by_property = ("order_by", ordering.order_by)
    pagination_dump = {key: str(value) for key, value in pagination.model_dump(exclude_defaults=True).items()}
    pagination_property = ("page", LogProperty("", children=pagination_dump) if pagination_dump else None)
    properties: Properties = {
        key: value for key, value in (filter_property, order_by_property, pagination_property) if value
    }
    return properties or None
