#
# Copyright (C) 2019-2025  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test admin REST API endpoints."""

from datetime import date, datetime, timezone
from typing import Any
from unittest.mock import ANY
from uuid import UUID

from django.test import TestCase, override_settings
from djmoney.money import Money
from freezegun import freeze_time
from grill import LogProperty

from django_pain.api.schema import TokenData
from django_pain.api.token import create_token
from django_pain.constants import (
    ApiPermissionScope,
    InvoiceType,
    LogEntryType,
    LogReferenceType,
    LogResult,
    PaymentState,
)
from django_pain.models import BankAccount, InvoiceRef, PaymentImportHistory
from django_pain.models.invoices import BillingItemType, InvoiceNumberSequence, VariableSymbolSequence, VatRate
from django_pain.tests.mixins import CacheResetMixin
from django_pain.tests.utils import (
    DummyPaymentProcessor,
    LoggerMixin,
    create_invoice,
    create_payment_request,
    get_account,
    get_client_ref,
    get_customer,
    get_invoice,
    get_payment,
    get_payment_request,
)


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestBankAccount(LoggerMixin, TestCase):
    """Tests of bank account endpoints."""

    # --- get detail ---

    def test_get_bank_account_info_not_found(self):
        response = self.client.get("/api/admin/bank_account/no-i-do-not-exist/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Bank account 'no-i-do-not-exist' not found"})
        self.assertLog(
            LogEntryType.ACCOUNT_DETAIL,
            LogResult.FAIL,
            input_properties={"account_id": "no-i-do-not-exist"},
        )

    def test_get_bank_account_info_ok(self):
        bank_account = BankAccount.objects.create(account_number="123456", account_name="Test account", currency="CZK")

        response = self.client.get(f"/api/admin/bank_account/{bank_account.uuid}/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "uid": str(bank_account.uuid),
            "number": "123456",
            "name": "Test account",
            "currency": "CZK",
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(
            LogEntryType.ACCOUNT_DETAIL,
            LogResult.SUCCESS,
            input_properties={"account_id": str(bank_account.uuid)},
            output_references={LogReferenceType.ACCOUNT: str(bank_account.uuid)},
        )

    # --- create new ---

    def test_create_bank_account_bad_param(self):
        response = self.client.post("/api/admin/bank_account/", data={}, content_type="application/json")

        self.assertEqual(response.status_code, 422)  # parameter validation error

    def test_create_bank_account_already_exists(self):
        bank_account = get_account(account_number="123456", account_name="", currency="CZK")
        bank_account.save()
        post_data = {"number": "123456", "currency": "EUR"}

        response = self.client.post("/api/admin/bank_account/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Bank account '123456' already exists"})
        self.assertLog(
            LogEntryType.ACCOUNT_CREATE,
            LogResult.FAIL,
            input_properties={"number": "123456", "currency": "EUR"},
        )

    def test_create_bank_account_ok(self):
        post_data = {"number": "123456", "name": "Test account", "currency": "CZK"}

        response = self.client.post("/api/admin/bank_account/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 201)
        post_data.update({"uid": ANY})
        self.assertJSONEqual(response.content, post_data)
        self.assertQuerySetEqual(
            BankAccount.objects.values_list("account_number", "account_name", "currency"),
            [("123456", "Test account", "CZK")],
            transform=tuple,
            ordered=False,
        )
        self.assertLog(
            LogEntryType.ACCOUNT_CREATE,
            LogResult.SUCCESS,
            input_properties={"number": "123456", "name": "Test account", "currency": "CZK"},
            output_references={LogReferenceType.ACCOUNT: ANY},
        )

    # --- delete existing ---

    def test_delete_bank_account_not_found(self):
        # This also tests proper passing a parameter with slashes
        response = self.client.delete("/api/admin/bank_account/i-do-not-exist/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Bank account 'i-do-not-exist' not found"})
        self.assertLog(
            LogEntryType.ACCOUNT_DELETE,
            LogResult.FAIL,
            input_properties={"account_id": "i-do-not-exist"},
        )

    def test_delete_bank_account_ok(self):
        bank_account = BankAccount.objects.create(account_number="123456", account_name="", currency="CZK")

        response = self.client.delete(f"/api/admin/bank_account/{bank_account.uuid}/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertIsNone(BankAccount.objects.filter(account_number="123456").first())
        self.assertLog(
            LogEntryType.ACCOUNT_DELETE,
            LogResult.SUCCESS,
            input_properties={"account_id": str(bank_account.uuid)},
            output_references={LogReferenceType.ACCOUNT: str(bank_account.uuid)},
        )

    def test_delete_bank_account_error(self):
        # Create account referenced by foreign key from other object (with RESTRICT constraint)
        bank_account = BankAccount.objects.create(account_number="123456", account_name="", currency="CZK")
        customer = get_customer()
        customer.save()
        payment_request = get_payment_request(bank_account=bank_account, customer=customer)
        payment_request.save()

        response = self.client.delete(f"/api/admin/bank_account/{bank_account.uuid}/")

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(
            response.content, {"details": f"Bank account '{bank_account.uuid}' not deleted because it is in use"}
        )
        self.assertLog(
            LogEntryType.ACCOUNT_DELETE,
            LogResult.FAIL,
            input_properties={"account_id": str(bank_account.uuid)},
        )

    # --- update existing ---

    def test_update_bank_account_bad_param(self):
        response = self.client.put("/api/admin/bank_account/number/", data={}, content_type="application/json")

        self.assertEqual(response.status_code, 422)  # parameter validation error

    def test_update_bank_account_not_found(self):
        post_data = {"name": "Reserve"}

        response = self.client.put("/api/admin/bank_account/123456/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Bank account '123456' not found"})
        self.assertLog(
            LogEntryType.ACCOUNT_UPDATE,
            LogResult.FAIL,
            input_properties={"account_id": "123456", "name": "Reserve"},
        )

    def test_update_bank_account_ok(self):
        bank_account = BankAccount.objects.create(account_number="123456", account_name="", currency="CZK")
        post_data = {"name": "Reserve"}

        response = self.client.put(
            f"/api/admin/bank_account/{bank_account.uuid}/", data=post_data, content_type="application/json"
        )

        self.assertEqual(response.status_code, 200)
        expected = {"uid": str(bank_account.uuid), "number": "123456", "name": "Reserve", "currency": "CZK"}
        self.assertJSONEqual(response.content, expected)
        self.assertQuerySetEqual(
            BankAccount.objects.values_list("account_number", "account_name", "currency"),
            [("123456", "Reserve", "CZK")],
            transform=tuple,
            ordered=False,
        )
        self.assertLog(
            LogEntryType.ACCOUNT_UPDATE,
            LogResult.SUCCESS,
            input_properties={"account_id": str(bank_account.uuid), "name": "Reserve"},
            output_references={LogReferenceType.ACCOUNT: str(bank_account.uuid)},
        )

    # --- list items ---

    def test_list_bank_accounts_no_params(self):
        bank_account = get_account(account_number="123456", account_name="", currency="CZK")
        bank_account.save()

        response = self.client.get("/api/admin/bank_accounts/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                {"uid": ANY, "number": "123456", "name": "", "currency": "CZK"},
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.ACCOUNT_LIST, LogResult.SUCCESS)

    def test_list_bank_accounts_no_items(self):
        response = self.client.get("/api/admin/bank_accounts/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [],
            "count": 0,
        }
        self.assertJSONEqual(response.content, expected)

    def test_list_bank_accounts_filter_by_number(self):
        bank_account = get_account(account_number="123456", account_name="", currency="CZK")
        bank_account.save()
        bank_account = get_account(account_number="ABCDEF", account_name="", currency="CZK")
        bank_account.save()

        response = self.client.get("/api/admin/bank_accounts/", {"number": "345"})

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                {"uid": ANY, "number": "123456", "name": "", "currency": "CZK"},
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(
            LogEntryType.ACCOUNT_LIST,
            LogResult.SUCCESS,
            input_properties={"filter": LogProperty("", children={"number": "345"})},
        )

    def test_list_bank_accounts_filter_by_name(self):
        bank_account = get_account(account_number="123456", account_name="Account 1", currency="CZK")
        bank_account.save()
        bank_account = get_account(account_number="ABCDEF", account_name="Account 2", currency="CZK")
        bank_account.save()

        response = self.client.get("/api/admin/bank_accounts/", {"name": "1"})

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                {"uid": ANY, "number": "123456", "name": "Account 1", "currency": "CZK"},
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)

    def test_list_bank_accounts_filter_by_currency(self):
        bank_account = get_account(account_number="123456", account_name="", currency="CZK")
        bank_account.save()
        bank_account = get_account(account_number="234567", account_name="", currency="EUR")
        bank_account.save()

        response = self.client.get("/api/admin/bank_accounts/", {"currency": "czk"})

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                {"uid": ANY, "number": "123456", "name": "", "currency": "CZK"},
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)

    def test_list_bank_accounts_filter_by_all(self):
        bank_account = get_account(account_number="123456", account_name="Account 1", currency="CZK")
        bank_account.save()
        bank_account = get_account(account_number="ABCDEF", account_name="Account 2", currency="EUR")
        bank_account.save()
        filters = {"number": "123", "name": "1", "currency": "czk"}

        response = self.client.get("/api/admin/bank_accounts/", filters)

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                {"uid": ANY, "number": "123456", "name": "Account 1", "currency": "CZK"},
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)

    def _create_account_testset(self):
        # Testing data for pagination and ordering
        data = [
            ("1", "B", "EUR"),
            ("2", "A", "CZK"),
            ("3", "E", "CZK"),
            ("4", "C", "CZK"),
            ("5", "D", "EUR"),
        ]
        for number, name, currency in data:
            BankAccount.objects.create(account_number=number, account_name=name, currency=currency)

    def test_list_bank_accounts_order(self):
        self._create_account_testset()
        params = {"order_by": "-currency,+name"}

        response = self.client.get("/api/admin/bank_accounts/", params)

        self.assertEqual(response.status_code, 200)
        expected = [
            {"uid": ANY, "number": "1", "name": "B", "currency": "EUR"},
            {"uid": ANY, "number": "5", "name": "D", "currency": "EUR"},
            {"uid": ANY, "number": "2", "name": "A", "currency": "CZK"},
            {"uid": ANY, "number": "4", "name": "C", "currency": "CZK"},
            {"uid": ANY, "number": "3", "name": "E", "currency": "CZK"},
        ]
        self.assertEqual(response.json()["items"], expected)
        self.assertEqual(response.json()["count"], 5)
        self.assertLog(LogEntryType.ACCOUNT_LIST, LogResult.SUCCESS, input_properties={"order_by": "-currency,+name"})


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestBankPayment(CacheResetMixin, LoggerMixin, TestCase):
    """Tests of bank payment endpoints."""

    @freeze_time("2024-10-01T09:00:00Z")
    def setUp(self):
        super().setUp()
        self.account = BankAccount.objects.create(account_number="123456/0100", currency="CZK")
        self.payment = get_payment(
            account=self.account,
            transaction_date=date(2024, 9, 30),
        )
        self.payment.save()

    # --- get detail ---

    def test_get_bank_payment_info_invalid_uuid(self):
        response = self.client.get("/api/admin/bank_payment/invalid-uuid/")

        self.assertEqual(response.status_code, 404)

    def test_get_bank_payment_info_not_found(self):
        response = self.client.get("/api/admin/bank_payment/00000000-0000-0000-0000-000000000001/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(
            response.content,
            {"details": "Bank payment '00000000-0000-0000-0000-000000000001' not found"},
        )
        self.assertLog(
            LogEntryType.PAYMENT_DETAIL,
            LogResult.FAIL,
            input_properties={"payment_id": "00000000-0000-0000-0000-000000000001"},
        )

    def test_get_bank_payment_info_ok(self):
        response = self.client.get(f"/api/admin/bank_payment/{self.payment.uuid}/")

        self.assertEqual(response.status_code, 200)
        expected: dict[str, Any] = {
            "uuid": str(self.payment.uuid),
            "identifier": "PAYMENT1",
            "account_number": "123456/0100",
            "counter_account_number": "098765/4321",
            "counter_account_name": "Another account",
            "amount": "42.0000000000",
            "currency": "CZK",
            "transaction_date": "2024-09-30",
            "constant_symbol": "",
            "variable_symbol": "",
            "specific_symbol": "",
            "description": "",
            "created": "2024-10-01T09:00:00Z",
            "payment_type": "transfer",
            "state": "ready_to_process",
            "processor": "",
            "processing_error": None,
            "card_payment_state": "",
            "card_handler": "",
            "custom_expiry": None,
            "customer_code": "",
            "email": None,
            "email_sent": None,
            "client": None,
            "invoices": [],
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(
            LogEntryType.PAYMENT_DETAIL,
            LogResult.SUCCESS,
            input_properties={"payment_id": str(self.payment.uuid)},
            output_references={LogReferenceType.PAYMENT: str(self.payment.uuid)},
        )

    def test_get_bank_payment_info_with_invoices(self):
        data = [
            ("500-1", 501, InvoiceType.ACCOUNT),
            ("500-2", 502, InvoiceType.ADVANCE),
        ]
        for number, remote_id, invoice_type in data:
            invoice = InvoiceRef.objects.create(number=number, remote_id=remote_id, invoice_type=invoice_type)
            invoice.payments.add(self.payment)

        response = self.client.get(f"/api/admin/bank_payment/{self.payment.uuid}/")

        self.assertEqual(response.status_code, 200)

        expected = [
            {"number": "500-1", "file_uid": None, "invoice_type": "account"},
            {"number": "500-2", "file_uid": None, "invoice_type": "advance"},
        ]
        self.assertEqual(response.json()["invoices"], expected)

    def test_get_bank_payment_info_with_client(self):
        client_ref = get_client_ref(payment=self.payment)
        client_ref.save()

        response = self.client.get(f"/api/admin/bank_payment/{self.payment.uuid}/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["client"], {"handle": "HANDLE"})

    # --- mark as refunded ---

    def test_mark_bank_payment_refunded_not_found(self):
        response = self.client.post("/api/admin/bank_payment/not-found/refund/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(
            response.content,
            {"details": "Bank payment 'not-found' not found"},
        )
        self.assertLog(
            LogEntryType.PAYMENT_REFUNDED,
            LogResult.FAIL,
            input_properties={"payment_id": "not-found"},
        )

    def test_mark_bank_payment_refunded_error(self):
        response = self.client.post(f"/api/admin/bank_payment/{self.payment.uuid}/refund/")

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(
            response.content,
            {"details": f"Bank payment '{self.payment.uuid}' not ready for refunding"},
        )
        self.assertLog(
            LogEntryType.PAYMENT_REFUNDED,
            LogResult.FAIL,
            input_properties={"payment_id": str(self.payment.uuid)},
        )

    def test_mark_bank_payment_refunded_to_refund_ok(self):
        self.payment.state = PaymentState.TO_REFUND
        self.payment.save(update_fields=["state"])

        response = self.client.post(f"/api/admin/bank_payment/{self.payment.uuid}/refund/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["state"], "refunded")
        self.assertLog(
            LogEntryType.PAYMENT_REFUNDED,
            LogResult.SUCCESS,
            input_properties={"payment_id": str(self.payment.uuid)},
            output_references={LogReferenceType.PAYMENT: str(self.payment.uuid)},
        )

    def test_mark_bank_payment_refunded_deferred_ok(self):
        self.payment.state = PaymentState.DEFERRED
        self.payment.save(update_fields=["state"])

        response = self.client.post(f"/api/admin/bank_payment/{self.payment.uuid}/refund/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["state"], "refunded")

    # --- assign payment ---

    def test_assign_payment_invalid_processor(self):
        data = {"processor": "unknown"}

        response = self.client.post(
            "/api/admin/bank_payment/00000000-0000-0000-0000-000000000001/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 422)

    @override_settings(PAIN_PROCESSORS={"success": "django_pain.tests.admin.test_forms.SuccessPaymentProcessor"})
    def test_assign_payment_missing_tax_date(self):
        data = {"processor": "success"}

        response = self.client.post(
            "/api/admin/bank_payment/00000000-0000-0000-0000-000000000001/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 422)

    @override_settings(PAIN_PROCESSORS={"success": "django_pain.tests.admin.test_forms.SuccessPaymentProcessor"})
    def test_assign_payment_not_found(self):
        data = {"processor": "success", "tax_date": "2024-10-01"}

        response = self.client.post(
            "/api/admin/bank_payment/not-found/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(
            response.content,
            {"details": "Bank payment 'not-found' not found"},
        )
        self.assertLog(
            LogEntryType.PAYMENT_ASSIGN,
            LogResult.FAIL,
            input_properties={
                "payment_id": "not-found",
                "processor": "success",
                "tax_date": "2024-10-01",
            },
        )

    @override_settings(PAIN_PROCESSORS={"success": "django_pain.tests.admin.test_forms.SuccessPaymentProcessor"})
    def test_assign_payment_bad_state(self):
        self.payment.state = PaymentState.CANCELED
        self.payment.save(update_fields=["state"])
        data = {"processor": "success", "tax_date": "2024-10-01"}

        response = self.client.post(
            f"/api/admin/bank_payment/{self.payment.uuid}/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(
            response.content,
            {"details": "Bank payment in state 'canceled' - not bound for processing"},
        )
        self.assertLog(
            LogEntryType.PAYMENT_ASSIGN,
            LogResult.FAIL,
            input_properties={
                "payment_id": str(self.payment.uuid),
                "processor": "success",
                "tax_date": "2024-10-01",
            },
        )

    @override_settings(PAIN_PROCESSORS={"success": "django_pain.tests.admin.test_forms.SuccessPaymentProcessor"})
    def test_assign_payment_processed(self):
        data = {"processor": "success", "tax_date": "2024-10-01"}

        response = self.client.post(
            f"/api/admin/bank_payment/{self.payment.uuid}/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 200)
        self.payment.refresh_from_db()
        self.assertEqual(self.payment.processor, "success")
        self.assertEqual(self.payment.state, PaymentState.PROCESSED)
        self.assertLog(
            LogEntryType.PAYMENT_ASSIGN,
            LogResult.SUCCESS,
            input_properties={
                "payment_id": str(self.payment.uuid),
                "processor": "success",
                "tax_date": "2024-10-01",
            },
            output_references={LogReferenceType.PAYMENT: str(self.payment.uuid)},
        )

    @override_settings(PAIN_PROCESSORS={"to_refund": "django_pain.tests.admin.test_forms.ToRefundPaymentProcessor"})
    def test_assign_payment_to_refund(self):
        data = {"processor": "to_refund", "hint": "do-it"}

        response = self.client.post(
            f"/api/admin/bank_payment/{self.payment.uuid}/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 200)
        self.payment.refresh_from_db()
        self.assertEqual(self.payment.processor, "to_refund")
        self.assertEqual(self.payment.state, PaymentState.TO_REFUND)
        self.assertLog(
            LogEntryType.PAYMENT_ASSIGN,
            LogResult.SUCCESS,
            input_properties={"payment_id": str(self.payment.uuid), "processor": "to_refund", "hint": "do-it"},
            output_references={LogReferenceType.PAYMENT: str(self.payment.uuid)},
        )

    @override_settings(PAIN_PROCESSORS={"failure": "django_pain.tests.admin.test_forms.FailurePaymentProcessor"})
    def test_assign_payment_not_processed(self):
        data = {"processor": "failure"}

        response = self.client.post(
            f"/api/admin/bank_payment/{self.payment.uuid}/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(
            response.content,
            {"details": "Payment not processed by 'failure' processor"},
        )
        self.assertLog(
            LogEntryType.PAYMENT_ASSIGN,
            LogResult.FAIL,
            input_properties={"payment_id": str(self.payment.uuid), "processor": "failure"},
        )

    @override_settings(PAIN_PROCESSORS={"bad_tax_date": "django_pain.tests.admin.test_forms.ExceptionPaymentProcessor"})
    def test_assign_payment_tax_date_exception(self):
        data = {"processor": "bad_tax_date"}

        response = self.client.post(
            f"/api/admin/bank_payment/{self.payment.uuid}/assign/",
            data=data,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Invalid tax date for assignment"})
        self.assertLog(
            LogEntryType.PAYMENT_ASSIGN,
            LogResult.FAIL,
            input_properties={"payment_id": str(self.payment.uuid), "processor": "bad_tax_date"},
        )

    # --- list payments ---

    def test_list_payments(self):
        response = self.client.get("/api/admin/bank_payments/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                {
                    "uuid": str(self.payment.uuid),
                    "identifier": "PAYMENT1",
                    "account_number": "123456/0100",
                    "counter_account_number": "098765/4321",
                    "counter_account_name": "Another account",
                    "amount": "42.0000000000",
                    "currency": "CZK",
                    "transaction_date": "2024-09-30",
                    "constant_symbol": "",
                    "variable_symbol": "",
                    "specific_symbol": "",
                    "description": "",
                    "created": "2024-10-01T09:00:00Z",
                    "payment_type": "transfer",
                    "state": "ready_to_process",
                    "processor": "",
                    "processing_error": None,
                    "card_payment_state": "",
                    "card_handler": "",
                    "custom_expiry": None,
                    "customer_code": "",
                    "email": None,
                    "email_sent": None,
                    "client": None,
                    "invoices": [],
                }
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.PAYMENT_LIST, LogResult.SUCCESS)

    def test_list_payments_with_filters(self):
        self.payment.variable_symbol = "24001"
        self.payment.save(update_fields=["raw_variable_symbol"])
        query_params = {
            "identifier": "PAYMENT1",
            "account_number": "123456/0100",
            "counter_account_number": "098765/4321",
            "amount": "42",
            "currency": "czk",
            "variable_symbol": "24001",
            "state": "ready_to_process",
        }

        response = self.client.get("/api/admin/bank_payments/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_payments_filter_off(self):
        self.payment.variable_symbol = "24001"
        self.payment.save(update_fields=["raw_variable_symbol"])
        data: list[dict[str, str]] = [
            {"identifier": "PAYMENT2"},
            {"account_number": "520/0200"},
            {"counter_account_number": "521/0300"},
            {"amount": "100"},
            {"currency": "usd"},
            {"variable_symbol": "24002"},
            {"state": "initialized"},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/bank_payments/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # payment filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_list_payments_filter_vs(self):
        self.payment.variable_symbol = "00123"
        self.payment.save(update_fields=["raw_variable_symbol"])
        data = [  # (trim_varsym, pattern, count)
            # PAIN_TRIM_VARSYM = False => "00123" in JSON response
            (False, "00123", 1),  # exact match
            (False, "123", 1),  # full info match
            (False, "23", 0),  # partial info only
            # PAIN_TRIM_VARSYM = True => "123" in JSON response
            (True, "00123", 1),  # match with stored value
            (True, "123", 1),  # match with showed value
            (True, "23", 0),  # partial info only
        ]
        for trim_varsym, pattern, count in data:
            with self.subTest(trim_varsym=trim_varsym, pattern=pattern):
                with override_settings(PAIN_TRIM_VARSYM=trim_varsym):
                    query_params = {"variable_symbol": pattern}

                    response = self.client.get("/api/admin/bank_payments/", query_params)

                    self.assertEqual(response.status_code, 200)
                    self.assertEqual(response.json()["count"], count)

    def test_list_payments_filter_transaction_date(self):
        data = [  # our payment has tranaction date 2024-09-30
            ({"transaction_date_from": "2024-09-29"}, 1),
            ({"transaction_date_from": "2024-09-29", "transaction_date_to": "2024-10-15"}, 1),
            ({"transaction_date_from": "2024-09-30", "transaction_date_to": "2024-10-15"}, 1),
            ({"transaction_date_from": "2024-10-01", "transaction_date_to": "2024-10-15"}, 0),
            ({"transaction_date_from": "2024-09-01", "transaction_date_to": "2024-09-29"}, 0),
            ({"transaction_date_from": "2024-09-01", "transaction_date_to": "2024-09-30"}, 1),
            ({"transaction_date_from": "2024-09-01", "transaction_date_to": "2024-10-01"}, 1),
            ({"transaction_date_to": "2024-10-01"}, 1),
        ]
        for query_params, count in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/bank_payments/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["count"], count)

    def test_list_payments_filter_multiple_state(self):
        data = [  # our payment is `ready_to_process`
            ("state=initialized&state=canceled", 0),
            ("state=ready_to_process&state=refunded&state=to_refund", 1),
        ]
        for query_params, count in data:
            with self.subTest(query_params=query_params):
                response = self.client.get(f"/api/admin/bank_payments/?{query_params}")

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["count"], count)

    def test_list_payments_ordering(self):
        self.payment.variable_symbol = "24001"
        self.payment.save(update_fields=["raw_variable_symbol"])
        account = get_account(account_number="9876/5000", currency="USD")
        account.save()
        with freeze_time("2024-10-02T10:00:00Z"):
            payment = get_payment(
                account=account,
                identifier="PAYMENT2",
                counter_account_number="100/0100",
                amount=Money(51.0, "USD"),
                transaction_date=date(2024, 10, 1),
                variable_symbol="24002",
                state=PaymentState.TO_REFUND,
            )
            payment.save()  # greater in all fields used in ordering

        for param in (
            "identifier",
            "account_number",
            "counter_account_number",
            "amount",
            "currency",
            "variable_symbol",
            "transaction_date",
            "state",
            "created",
        ):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/bank_payments/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["identifier"], "PAYMENT1")
                self.assertEqual(response.json()["items"][1]["identifier"], "PAYMENT2")
                # descending order
                response = self.client.get("/api/admin/bank_payments/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["identifier"], "PAYMENT2")
                self.assertEqual(response.json()["items"][1]["identifier"], "PAYMENT1")


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestPaymentImportHistory(LoggerMixin, TestCase):
    """Tests of payment import history endpoints."""

    @freeze_time("2024-10-01T12:00:00Z")
    def setUp(self):
        super().setUp()
        self.payment_import_history = PaymentImportHistory.objects.create(
            origin="bank1",
            errors=0,
            finished=True,
        )

    # --- get detail ---

    def test_get_payment_import_history_info_not_found(self):
        response = self.client.get("/api/admin/payment_import_history/not-found/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(
            response.content,
            {"details": "Payment import history item 'not-found' not found"},
        )
        self.assertLog(LogEntryType.IMPORT_DETAIL, LogResult.FAIL, input_properties={"import_id": "not-found"})

    def test_payment_import_history_info_ok(self):
        response = self.client.get(f"/api/admin/payment_import_history/{self.payment_import_history.uuid}/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "uuid": str(self.payment_import_history.uuid),
            "origin": "bank1",
            "start_datetime": "2024-10-01T12:00:00Z",
            "filenames": [],
            "errors": 0,
            "finished": True,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(
            LogEntryType.IMPORT_DETAIL,
            LogResult.SUCCESS,
            input_properties={"import_id": str(self.payment_import_history.uuid)},
        )

    def test_payment_import_history_info_ok_with_filenames(self):
        self.payment_import_history._filenames = PaymentImportHistory.separator.join(("file1", "file2", "file3"))
        self.payment_import_history.save(update_fields=["_filenames"])

        response = self.client.get(f"/api/admin/payment_import_history/{self.payment_import_history.uuid}/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["filenames"], ["file1", "file2", "file3"])

    # --- list items ---

    def test_list_payment_import_history(self):
        response = self.client.get("/api/admin/payment_import_history/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                {
                    "uuid": str(self.payment_import_history.uuid),
                    "origin": "bank1",
                    "start_datetime": "2024-10-01T12:00:00Z",
                    "filenames": [],
                    "errors": 0,
                    "finished": True,
                }
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.IMPORT_LIST, LogResult.SUCCESS)

    def test_list_payment_import_history_with_filters(self):
        self.payment_import_history._filenames = "file1"
        self.payment_import_history.save(update_fields=["_filenames"])
        query_params = {
            "origin": "bank1",
            "filenames": "le1",
            "finished": True,
        }

        response = self.client.get("/api/admin/payment_import_history/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_payment_import_history_filter_off(self):
        data: list[dict[str, str]] = [
            {"origin": "bank2"},
            {"filenames": "file1"},
            {"finished": "false"},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/payment_import_history/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_list_payment_import_history_filter_start_date(self):
        data = [  # our record has start 2024-10-01T12:00:00Z
            ({"start_datetime_from": "2024-09-29T00:00Z"}, 1),
            ({"start_datetime_from": "2024-09-30T00:00Z", "start_datetime_to": "2024-10-15T00:00Z"}, 1),
            ({"start_datetime_from": "2024-10-01T12:00Z", "start_datetime_to": "2024-10-15T00:00Z"}, 1),
            ({"start_datetime_from": "2024-10-02T00:00Z", "start_datetime_to": "2024-10-15T00:00Z"}, 0),
            ({"start_datetime_from": "2024-09-01T00:00Z", "start_datetime_to": "2024-10-01T12:00Z"}, 0),
            ({"start_datetime_from": "2024-09-01T00:00Z", "start_datetime_to": "2024-10-01T20:00Z"}, 1),
            ({"start_datetime_from": "2024-09-01T00:00Z", "start_datetime_to": "2024-10-02T20:00Z"}, 1),
            ({"start_datetime_to": "2024-10-02T20:00Z"}, 1),
        ]
        for query_params, count in data:
            with self.subTest(quary_params=query_params):
                response = self.client.get("/api/admin/payment_import_history/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["count"], count)

    @freeze_time("2024-10-01T10:00:00Z")
    def test_list_payment_import_history_ordering(self):
        payment_import_history = PaymentImportHistory.objects.create(
            origin="bank0",
            errors=0,
            finished=False,
        )
        payment_import_history.save()  # less in all fields used in ordering

        for param in ("origin", "start_datetime"):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/payment_import_history/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["origin"], "bank0")
                self.assertEqual(response.json()["items"][1]["origin"], "bank1")
                # descending order
                response = self.client.get("/api/admin/payment_import_history/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["origin"], "bank1")
                self.assertEqual(response.json()["items"][1]["origin"], "bank0")


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestCustomer(LoggerMixin, TestCase):
    """Tests of customer endpoints."""

    def setUp(self):
        super().setUp()
        customer = get_customer(
            uid="cid-1",
            email="test@example.org",
            phone_number="+420 100 200 300",
            vies_vat_valid=False,
            vies_checked_at=datetime(2024, 10, 1, tzinfo=timezone.utc),
        )
        customer.save()
        self.expected_item_response = {
            "uid": "cid-1",
            "name": None,
            "organization": "Test, s.r.o.",
            "street": ["Na Okraji 1"],
            "city": "Pokuston",
            "state_or_province": None,
            "postal_code": "050 10",
            "country": {"code": "CZ", "name": "Czechia"},
            "email": "test@example.org",
            "phone_number": "+420 100 200 300",
            "company_registration_number": "001234",
            "vat_number": "CZ001234",
            "vies_vat_valid": False,
            "vies_checked_at": "2024-10-01T00:00:00Z",
        }

    # --- get detail ---

    def test_customer_info_not_found(self):
        response = self.client.get("/api/admin/customer/cid-2/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Customer 'cid-2' not found"})
        self.assertLog(
            LogEntryType.CUSTOMER_DETAIL,
            LogResult.FAIL,
            input_properties={"customer_id": "cid-2"},
        )

    def test_customer_info_ok(self):
        response = self.client.get("/api/admin/customer/cid-1/")

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, self.expected_item_response)
        self.assertLog(
            LogEntryType.CUSTOMER_DETAIL,
            LogResult.SUCCESS,
            input_properties={"customer_id": "cid-1"},
            output_references={LogReferenceType.CUSTOMER: "cid-1"},
        )

    # --- list items ---

    def test_list_customers_ok(self):
        response = self.client.get("/api/admin/customers/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                self.expected_item_response,
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.CUSTOMER_LIST, LogResult.SUCCESS)

    def test_list_customers_with_filters(self):
        query_params = {
            # address filters
            "name": "s.r.o.",
            "address": "okraj",
            "postal_code": "050 10",
            "country": "cz",
            "company_registration_number": "001234",
            "vat_number": "CZ001234",
            # customer specific filters
            "email": "test@example.org",
            "vies_vat_valid": "false",
        }

        response = self.client.get("/api/admin/customers/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_customers_filter_off(self):
        data: list[dict[str, str]] = [
            # address filters
            {"name": "a.s."},
            {"address": "nowhere"},
            {"postal_code": "050 11"},
            {"country": "fr"},
            {"company_registration_number": "505"},
            {"vat_number": "CZ505"},
            # customer specific filters
            {"email": "test@example.cz"},
            {"vies_vat_valid": "true"},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/customers/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_customer_ordering(self):
        customer = get_customer(  # greater in ordering fields
            uid="cid-2",
            organization="Z Test, a.s.",
            country_code="DE",
            company_registration_number="9876",
            vat_number="DE9876",
            vies_vat_valid=True,
            vies_checked_at=datetime(2024, 10, 15, tzinfo=timezone.utc),
        )
        customer.save()

        for param in (
            "organization",
            "country",
            "company_registration_number",
            "vat_number",
            "vies_vat_valid",
            "vies_checked_at",
        ):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/customers/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["uid"], "cid-1")
                self.assertEqual(response.json()["items"][1]["uid"], "cid-2")
                # descending order
                response = self.client.get("/api/admin/customers/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["uid"], "cid-2")
                self.assertEqual(response.json()["items"][1]["uid"], "cid-1")


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestPaymentRequest(LoggerMixin, TestCase):
    """Tests of payment request endpoints."""

    def setUp(self):
        super().setUp()
        self.expected_item_response = {
            "uid": "00000000-0000-0000-0000-000000000001",
            "customer_uid": "CID-1",
            "app_id": "auctions",
            "name": None,
            "organization": "Test, s.r.o.",
            "street": ["Na Okraji 1"],
            "city": "Pokuston",
            "state_or_province": None,
            "postal_code": "050 10",
            "country": {"code": "CZ", "name": "Czechia"},
            "company_registration_number": "001234",
            "vat_number": "CZ001234",
            "created_at": "2023-09-01T12:00:00Z",
            "due_at": None,
            "due_at_displayed": "2023-10-01T12:00:00Z",
            "closed_at": None,
            "price_to_pay": "74.0000000000",
            "currency": "CZK",
            "variable_symbol": "0000009876",
            "account": "123456/0300",
            "card_payment": None,
            "realized_payment": "00000000-0000-0000-0000-00000000001c",
            "file_uid": None,
            "invoice_uid": None,
            "items": [
                {
                    "name": "Item 1",
                    "item_type": "basic_rate",
                    "quantity": 1,
                    "price_per_unit": "11.0000000000",
                    "price_total": "11.0000000000",
                },
                {
                    "name": "Item 2",
                    "item_type": "basic_rate",
                    "quantity": 2,
                    "price_per_unit": "12.0000000000",
                    "price_total": "24.0000000000",
                },
                {
                    "name": "Item 3",
                    "item_type": "basic_rate",
                    "quantity": 3,
                    "price_per_unit": "13.0000000000",
                    "price_total": "39.0000000000",
                },
            ],
        }

    # --- get detail ---

    def test_payment_request_info_not_found(self):
        response = self.client.get("/api/admin/payment_request/not-found/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Payment request 'not-found' not found"})
        self.assertLog(
            LogEntryType.PAYMENT_REQUEST_DETAIL,
            LogResult.FAIL,
            input_properties={"prequest_id": "not-found"},
        )

    def test_payment_request_info_ok(self):
        create_payment_request()

        response = self.client.get("/api/admin/payment_request/00000000-0000-0000-0000-000000000001/")

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, self.expected_item_response)
        self.assertLog(
            LogEntryType.PAYMENT_REQUEST_DETAIL,
            LogResult.SUCCESS,
            input_properties={"prequest_id": "00000000-0000-0000-0000-000000000001"},
            output_references={LogReferenceType.PAYMENT_REQUEST: "00000000-0000-0000-0000-000000000001"},
        )

    # --- close payment request ---

    def test_close_payment_request_not_found(self):
        response = self.client.post("/api/admin/payment_request/not-found/close/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Payment request 'not-found' not found"})
        self.assertLog(
            LogEntryType.PAYMENT_REQUEST_CLOSE,
            LogResult.FAIL,
            input_properties={"prequest_id": "not-found"},
        )

    @freeze_time("2024-10-01T10:00:00Z")
    def test_close_payment_request_info_ok(self):
        payment_request = create_payment_request()

        response = self.client.post("/api/admin/payment_request/00000000-0000-0000-0000-000000000001/close/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["closed_at"], "2024-10-01T10:00:00Z")
        payment_request.refresh_from_db()
        self.assertEqual(payment_request.closed_at, datetime(2024, 10, 1, 10, 0, 0, tzinfo=timezone.utc))
        self.assertLog(
            LogEntryType.PAYMENT_REQUEST_CLOSE,
            LogResult.SUCCESS,
            input_properties={"prequest_id": "00000000-0000-0000-0000-000000000001"},
            output_references={LogReferenceType.PAYMENT_REQUEST: "00000000-0000-0000-0000-000000000001"},
        )

    def test_close_payment_request_info_already_closed(self):
        create_payment_request(closed_at=datetime(2024, 9, 20, tzinfo=timezone.utc))

        response = self.client.post("/api/admin/payment_request/00000000-0000-0000-0000-000000000001/close/")

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(
            response.content,
            {"details": "Payment request '00000000-0000-0000-0000-000000000001' is already closed"},
        )
        self.assertLog(
            LogEntryType.PAYMENT_REQUEST_CLOSE,
            LogResult.FAIL,
            input_properties={"prequest_id": "00000000-0000-0000-0000-000000000001"},
        )

    # --- list items ---

    def test_list_payment_requests_ok(self):
        create_payment_request()

        response = self.client.get("/api/admin/payment_requests/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                self.expected_item_response,
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.PAYMENT_REQUEST_LIST, LogResult.SUCCESS)

    def test_list_payment_requests_with_filters(self):
        customer = get_customer(email="test@example.org")
        customer.save()
        create_payment_request(customer=customer, closed_at=datetime(2024, 10, 1, tzinfo=timezone.utc))
        query_params = {
            # address filters - see customer test
            # payment request specific filters
            "app_id": "auctions",
            "customer_uid": "CID-1",
            "email": "test@example.org",
            "created_at_from": "2023-09-01T00:00Z",
            "created_at_to": "2023-09-02T00:00Z",
            "closed_at_from": "2024-10-01T00:00Z",
            "closed_at_to": "2024-10-02T00:00Z",
            "price_to_pay": "74",
            "currency": "czk",
            "variable_symbol": "0000009876",
            "state": "realized",
            "item_name": "item",
        }

        response = self.client.get("/api/admin/payment_requests/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_payment_requests_filter_off(self):
        create_payment_request(closed_at=datetime(2024, 10, 1, tzinfo=timezone.utc))
        data: list[dict[str, str]] = [
            # address filters - see customer test
            # payment request specific filters
            {"app_id": "other"},
            {"customer_uid": "CID-2"},
            {"email": "test@example.cz"},
            {"created_at_from": "2023-09-02T00:00Z"},
            {"created_at_to": "2023-09-01T00:00Z"},
            {"closed_at_from": "2024-10-01T00:01Z"},
            {"closed_at_to": "2024-10-01T00:00Z"},
            {"price_to_pay": "74.1"},
            {"currency": "eur"},
            {"variable_symbol": "9876"},
            {"state": "waiting"},
            {"state": "closed"},
            {"state": "overdue"},
            {"item_name": "item 4"},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/payment_requests/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_payment_request_ordering(self):
        create_payment_request(
            due_at=datetime(2023, 10, 15, tzinfo=timezone.utc),
            closed_at=datetime(2023, 10, 30, tzinfo=timezone.utc),
        )
        customer = get_customer(
            uid="cid-2",
            organization="Z Test, a.s.",
            country_code="DE",
            company_registration_number="9876",
            vat_number="DE9876",
        )
        customer.save()
        account = get_account(account_number="9876/5000", currency="EUR")
        account.save()
        payment_request = get_payment_request(  # greater in ordering fields
            uuid=UUID(int=2),
            app_id="other",
            customer=customer,
            price_to_pay=1200,
            currency="EUR",
            bank_account=account,
            variable_symbol="5000012345",
            due_at=datetime(2024, 2, 10, tzinfo=timezone.utc),
            due_at_displayed=datetime(2024, 2, 10, tzinfo=timezone.utc),
            closed_at=datetime(2024, 2, 10, tzinfo=timezone.utc),
        )
        with freeze_time("2024-02-01T12:00Z"):
            payment_request.save()

        for param in (
            "app_id",
            "organization",
            "country",
            "company_registration_number",
            "vat_number",
            "created_at",
            "due_at",
            "due_at_displayed",
            "closed_at",
            "price_to_pay",
            "currency",
            "account",
            "variable_symbol",
        ):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/payment_requests/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["uid"], "00000000-0000-0000-0000-000000000001")
                self.assertEqual(response.json()["items"][1]["uid"], "00000000-0000-0000-0000-000000000002")
                # descending order
                response = self.client.get("/api/admin/payment_requests/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["uid"], "00000000-0000-0000-0000-000000000002")
                self.assertEqual(response.json()["items"][1]["uid"], "00000000-0000-0000-0000-000000000001")


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestInvoice(LoggerMixin, TestCase):
    """Tests of invoice endpoints."""

    def setUp(self):
        super().setUp()
        customer = get_customer(email="test@example.org")
        customer.save()
        self.invoice = create_invoice(customer=customer)
        self.expected_item_response = {
            "number": "902400001",
            "uid": str(self.invoice.uuid),
            "customer_uid": "CID-1",
            "app_id": "auctions",
            "name": None,
            "organization": "Test, s.r.o.",
            "street": ["Na Okraji 1"],
            "city": "Pokuston",
            "state_or_province": None,
            "postal_code": "050 10",
            "country": {"code": "CZ", "name": "Czechia"},
            "company_registration_number": "001234",
            "vat_number": "CZ001234",
            "created_at": "2024-01-31T15:00:00Z",
            "due_at": "2024-01-25T12:00:00Z",
            "tax_at": "2024-01-20",
            "paid_at": "2024-01-20",
            "variable_symbol": "0902400012",
            "account": "123456/0300",
            "payment_type": "transfer",
            "price_total_base": "41.0000000000",
            "price_total_vat": "7.1000000000",
            "price_total": "48.1000000000",
            "price_to_pay": "0E-10",
            "currency": "CZK",
            "file_uid": None,
            "items": [
                {
                    "name": "Item 1",
                    "item_type": "basic_rate",
                    "vat_rate": "10.0000",
                    "quantity": 1,
                    "price_per_unit": "11.0000000000",
                    "price_total_base": "11.0000000000",
                    "price_total_vat": "1.1000000000",
                    "price_total": "12.1000000000",
                },
                {
                    "name": "Item 2",
                    "item_type": "basic_rate",
                    "vat_rate": "20.0000",
                    "quantity": 2,
                    "price_per_unit": "15.0000000000",
                    "price_total_base": "30.0000000000",
                    "price_total_vat": "6.0000000000",
                    "price_total": "36.0000000000",
                },
            ],
        }

    # --- get detail ---

    def test_invoice_info_not_found(self):
        response = self.client.get("/api/admin/invoice/not-found/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Invoice 'not-found' not found"})
        self.assertLog(
            LogEntryType.INVOICE_DETAIL,
            LogResult.FAIL,
            input_properties={"invoice_id": "not-found"},
        )

    def test_invoice_info_ok(self):
        response = self.client.get(f"/api/admin/invoice/{self.invoice.uuid}/")

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, self.expected_item_response)
        self.assertLog(
            LogEntryType.INVOICE_DETAIL,
            LogResult.SUCCESS,
            input_properties={"invoice_id": str(self.invoice.uuid)},
            output_references={LogReferenceType.INVOICE: str(self.invoice.uuid)},
        )

    # --- list items ---

    def test_list_invoices_ok(self):
        response = self.client.get("/api/admin/invoices/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                self.expected_item_response,
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.INVOICE_LIST, LogResult.SUCCESS)

    def test_list_invoices_with_filters(self):
        query_params = {
            # address filters - see customer test
            # invoice specific filters
            "number": "902400001",
            "app_id": "auctions",
            "customer_uid": "CID-1",
            "email": "test@example.org",
            "price_total": "48.1",
            "currency": "czk",
            "variable_symbol": "0902400012",
            "tax_at_from": "2024-01-15",
            "tax_at_to": "2024-01-25",
            "paid_at_from": "2024-01-15",
            "paid_at_to": "2024-01-25",
            "item_name": "item",
        }

        response = self.client.get("/api/admin/invoices/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_invoices_filter_off(self):
        data: list[dict[str, str]] = [
            # address filters - see customer test
            # invoice specific filters
            {"number": "902400002"},
            {"app_id": "other"},
            {"customer_uid": "CID-2"},
            {"email": "test@example.cz"},
            {"price_total": "20.5"},
            {"currency": "eur"},
            {"variable_symbol": "0902400013"},
            {"tax_at_from": "2024-01-25"},
            {"tax_at_to": "2024-01-15"},
            {"paid_at_from": "2024-01-25"},
            {"paid_at_to": "2024-01-15"},
            {"item_name": "item 4"},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/invoices/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_invoice_ordering(self):
        customer = get_customer(
            uid="cid-2",
            organization="Z Test, a.s.",
            country_code="DE",
            company_registration_number="9876",
            vat_number="DE9876",
        )
        customer.save()
        invoice = get_invoice(  # greater in ordering fields
            customer=customer,
            number="902400002",
            app_id="other",
            due_at=datetime(2024, 3, 15, tzinfo=timezone.utc),
            tax_at=date(2024, 3, 10),
            paid_at=date(2024, 3, 10),
            variable_symbol="0902400013",
            bank_account="9876/5000",
            price_total=1200.50,
            currency="EUR",
        )
        with freeze_time("2024-03-01T12:00Z"):
            invoice.save()

        for param in (
            "number",
            "app_id",
            "organization",
            "country",
            "company_registration_number",
            "vat_number",
            "created_at",
            "due_at",
            "tax_at",
            "paid_at",
            "price_total",
            "currency",
            "account",
            "variable_symbol",
        ):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/invoices/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["number"], "902400001")
                self.assertEqual(response.json()["items"][1]["number"], "902400002")
                # descending order
                response = self.client.get("/api/admin/invoices/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["number"], "902400002")
                self.assertEqual(response.json()["items"][1]["number"], "902400001")


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestVariableSymbol(LoggerMixin, TestCase):
    """Tests of variable symbol sequence endpoints."""

    def setUp(self):
        super().setUp()
        VariableSymbolSequence.objects.create(app_id="auctions", year=2023, prefix=6123, current_number=6)
        self.expected_item_response = {
            "app_id": "auctions",
            "year": 2023,
            "prefix": 6123,
            "current_number": 6,
        }

    # --- create new ---

    def test_create_variable_symbol_already_exists(self):
        post_data = {"app_id": "auctions", "year": 2023, "prefix": 6123}

        response = self.client.post("/api/admin/variable_symbol/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Variable symbol sequence already exists"})
        self.assertLog(
            LogEntryType.VARIABLE_SYMBOL_CREATE,
            LogResult.FAIL,
            input_properties={"app_id": "auctions", "year": 2023, "prefix": 6123},
        )

    def test_create_variable_symbol_ok(self):
        post_data = {"app_id": "auctions", "year": 2024, "prefix": 6124}

        response = self.client.post("/api/admin/variable_symbol/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 201)
        self.assertJSONEqual(
            response.content,
            {"app_id": "auctions", "year": 2024, "prefix": 6124, "current_number": 0},
        )
        self.assertQuerySetEqual(
            VariableSymbolSequence.objects.filter(year=2024).values_list("app_id", "year", "prefix", "current_number"),
            [("auctions", 2024, 6124, 0)],
            transform=tuple,
            ordered=False,
        )
        self.assertLog(
            LogEntryType.VARIABLE_SYMBOL_CREATE,
            LogResult.SUCCESS,
            input_properties={"app_id": "auctions", "year": 2024, "prefix": 6124},
        )

    # --- list items ---

    def test_list_variable_symbols_ok(self):
        response = self.client.get("/api/admin/variable_symbols/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                self.expected_item_response,
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.VARIABLE_SYMBOL_LIST, LogResult.SUCCESS)

    def test_list_variable_symbols_with_filters(self):
        query_params = {
            "app_id": "auctions",
            "year": 2023,
        }

        response = self.client.get("/api/admin/variable_symbols/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_variable_symbols_filter_off(self):
        data: list[dict[str, str]] = [
            {"app_id": "other"},
            {"year": "2020"},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/variable_symbols/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_list_variable_symbols_ordering(self):
        VariableSymbolSequence.objects.create(app_id="other", year=2024, prefix=6124, current_number=0)
        for param in ("app_id", "year"):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/variable_symbols/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["year"], 2023)
                self.assertEqual(response.json()["items"][1]["year"], 2024)
                # descending order
                response = self.client.get("/api/admin/variable_symbols/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["year"], 2024)
                self.assertEqual(response.json()["items"][1]["year"], 2023)


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestInvoiceNumber(LoggerMixin, TestCase):
    """Tests of invoice number sequence endpoints."""

    def setUp(self):
        super().setUp()
        InvoiceNumberSequence.objects.create(app_id="auctions", year=2023, prefix="4023", current_number=6)
        self.expected_item_response = {
            "app_id": "auctions",
            "year": 2023,
            "prefix": "4023",
            "current_number": 6,
        }

    # --- create new ---

    def test_create_invoice_number_already_exists(self):
        post_data = {"app_id": "auctions", "year": 2023, "prefix": "4023"}

        response = self.client.post("/api/admin/invoice_number/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Invoice number sequence already exists"})
        self.assertLog(
            LogEntryType.INVOICE_NUMBER_CREATE,
            LogResult.FAIL,
            input_properties={"app_id": "auctions", "year": 2023, "prefix": "4023"},
        )

    def test_create_invoice_number_prefix(self):
        data = [
            "",  # empty string
            "A001",  # letters not allowed
            "402",  # too short
            "40240",  # too long
        ]
        for prefix in data:
            with self.subTest(prefix=prefix):
                post_data = {"app_id": "auctions", "year": 2024, "prefix": prefix}

                response = self.client.post(
                    "/api/admin/invoice_number/", data=post_data, content_type="application/json"
                )

                self.assertEqual(response.status_code, 422)  # parameter validation error

    def test_create_invoice_number_ok(self):
        post_data = {"app_id": "auctions", "year": 2024, "prefix": "4024"}

        response = self.client.post("/api/admin/invoice_number/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 201)
        self.assertJSONEqual(
            response.content,
            {"app_id": "auctions", "year": 2024, "prefix": "4024", "current_number": 0},
        )
        self.assertQuerySetEqual(
            InvoiceNumberSequence.objects.filter(year=2024).values_list("app_id", "year", "prefix", "current_number"),
            [("auctions", 2024, "4024", 0)],
            transform=tuple,
            ordered=False,
        )
        self.assertLog(
            LogEntryType.INVOICE_NUMBER_CREATE,
            LogResult.SUCCESS,
            input_properties={"app_id": "auctions", "year": 2024, "prefix": "4024"},
        )

    # --- list items ---

    def test_list_invoice_numbers_ok(self):
        response = self.client.get("/api/admin/invoice_numbers/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                self.expected_item_response,
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.INVOICE_NUMBER_LIST, LogResult.SUCCESS)

    def test_list_invoice_numbers_with_filters(self):
        query_params = {
            "app_id": "auctions",
            "year": 2023,
        }

        response = self.client.get("/api/admin/invoice_numbers/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_invoice_numbers_filter_off(self):
        data: list[dict[str, str]] = [
            {"app_id": "other"},
            {"year": "2020"},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/invoice_numbers/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_list_invoice_numbers_ordering(self):
        InvoiceNumberSequence.objects.create(app_id="other", year=2024, prefix="4024", current_number=0)
        for param in ("app_id", "year"):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/invoice_numbers/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["year"], 2023)
                self.assertEqual(response.json()["items"][1]["year"], 2024)
                # descending order
                response = self.client.get("/api/admin/invoice_numbers/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["year"], 2024)
                self.assertEqual(response.json()["items"][1]["year"], 2023)


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestItemType(LoggerMixin, TestCase):
    """Tests of billing item type endpoints."""

    def setUp(self):
        super().setUp()
        BillingItemType.objects.create(handle="example", not_invoiceable=False, cpa_code="01.23")
        self.expected_item_response = {
            "handle": "example",
            "not_invoiceable": False,
            "cpa_code": "01.23",
        }

    # --- create new ---

    def test_create_item_type_already_exists(self):
        post_data = {"handle": "example", "not_invoiceable": True, "cpa_code": ""}

        response = self.client.post("/api/admin/item_type/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {"details": "Billing item type 'example' already exists"})
        self.assertLog(
            LogEntryType.ITEM_TYPE_CREATE,
            LogResult.FAIL,
            input_properties={"handle": "example", "not_invoiceable": True, "cpa_code": ""},
        )

    def test_create_item_type_ok(self):
        post_data = {"handle": "other", "not_invoiceable": True, "cpa_code": ""}

        response = self.client.post("/api/admin/item_type/", data=post_data, content_type="application/json")

        self.assertEqual(response.status_code, 201)
        self.assertJSONEqual(
            response.content,
            {"handle": "other", "not_invoiceable": True, "cpa_code": ""},
        )
        self.assertQuerySetEqual(
            BillingItemType.objects.filter(handle="other").values_list("handle", "not_invoiceable", "cpa_code"),
            [("other", True, "")],
            transform=tuple,
            ordered=False,
        )
        self.assertLog(
            LogEntryType.ITEM_TYPE_CREATE,
            LogResult.SUCCESS,
            input_properties={"handle": "other", "not_invoiceable": True, "cpa_code": ""},
        )

    # --- list items ---

    def test_list_item_types_ok(self):
        response = self.client.get("/api/admin/item_types/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                self.expected_item_response,
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.ITEM_TYPE_LIST, LogResult.SUCCESS)

    def test_list_item_types_with_filters(self):
        query_params = {
            "handle": "XAM",
            "not_invoiceable": False,
        }

        response = self.client.get("/api/admin/item_types/", query_params)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_invoice_numbers_filter_off(self):
        data: list[dict[str, Any]] = [
            {"handle": "other"},
            {"not_invoiceable": True},
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/item_types/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_list_invoice_numbers_ordering(self):
        BillingItemType.objects.create(handle="other", not_invoiceable=True, cpa_code="")
        for param in ("handle",):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/item_types/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["handle"], "example")
                self.assertEqual(response.json()["items"][1]["handle"], "other")
                # descending order
                response = self.client.get("/api/admin/item_types/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["handle"], "other")
                self.assertEqual(response.json()["items"][1]["handle"], "example")


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestVatRate(LoggerMixin, TestCase):
    """Tests of VAT rate endpoints."""

    def setUp(self):
        super().setUp()
        VatRate.objects.create(
            country_code="CZ",
            cpa_code="01.23",
            valid_from=datetime(2023, 1, 1, tzinfo=timezone.utc),
            valid_to=datetime(2024, 1, 1, tzinfo=timezone.utc),
            value=21.0,
        )
        self.expected_item_response = {
            "country": {"code": "CZ", "name": "Czechia"},
            "cpa_code": "01.23",
            "valid_from": "2023-01-01",
            "valid_to": "2023-12-31",
            "value": "21.0000",
        }

    # --- list items ---

    def test_list_vat_rates_ok(self):
        response = self.client.get("/api/admin/vat_rates/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "items": [
                self.expected_item_response,
            ],
            "count": 1,
        }
        self.assertJSONEqual(response.content, expected)
        self.assertLog(LogEntryType.VAT_RATE_LIST, LogResult.SUCCESS)

    def test_list_vat_rates_with_filters(self):
        query_params = {"country": "CZ"}
        data = [
            "2023-01-01",  # first date of validity
            "2023-06-01",  # within validity interval
            "2023-12-31",  # last date of validity
        ]

        for valid_on in data:
            query_params["valid_on"] = valid_on
            with self.subTest(valid_on=valid_on):
                response = self.client.get("/api/admin/vat_rates/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["count"], 1)  # all filters pass

    def test_list_vat_rates_filter_off(self):
        data: list[dict[str, str]] = [
            {"country": "DE"},
            {"valid_on": "2022-12-31"},  # before validity
            {"valid_on": "2024-01-01"},  # after validity
        ]
        for query_params in data:
            with self.subTest(query_params=query_params):
                response = self.client.get("/api/admin/vat_rates/", query_params)

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json()["items"], [])  # record filtered off
                self.assertEqual(response.json()["count"], 0)

    def test_list_vat_rates_ordering(self):
        VatRate.objects.create(
            country_code="DE",
            cpa_code="45",
            valid_from=datetime(2023, 1, 1, tzinfo=timezone.utc),
            valid_to=None,
            value=20.0,
        )
        for param in ("country", "cpa_code"):
            with self.subTest(param=param):
                # ascending order
                response = self.client.get("/api/admin/vat_rates/", {"order_by": f"+{param}"})
                self.assertEqual(response.json()["items"][0]["country"]["code"], "CZ")
                self.assertEqual(response.json()["items"][1]["country"]["code"], "DE")
                # descending order
                response = self.client.get("/api/admin/vat_rates/", {"order_by": f"-{param}"})
                self.assertEqual(response.json()["items"][0]["country"]["code"], "DE")
                self.assertEqual(response.json()["items"][1]["country"]["code"], "CZ")


@override_settings(
    ROOT_URLCONF="django_pain.tests.urls",
    PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}},
)
class TestAuthentication(TestCase):
    """Tests of authentication with encrypted JWT."""

    # Notes:
    #  - Tests in the above classes pass because JWT authentication is not configured there.
    #  - Endpoint to update bank account is used as a respresentative for authentication tests.

    def setUp(self):
        self.bank_account = BankAccount.objects.create(account_number="123456", account_name="", currency="CZK")

    def test_authentication_fail_no_token(self):
        post_data = {"name": "Reserve"}

        response = self.client.put(
            f"/api/admin/bank_account/{self.bank_account.uuid}/", data=post_data, content_type="application/json"
        )

        self.assertEqual(response.status_code, 401)
        self.assertJSONEqual(response.content, {"detail": "Unauthorized"})

    def test_authentication_fail_empty_token(self):
        headers = {"Authorization": "Bearer"}
        post_data = {"name": "Reserve"}

        response = self.client.put(
            f"/api/admin/bank_account/{self.bank_account.uuid}/",
            data=post_data,
            headers=headers,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 401)
        self.assertJSONEqual(response.content, {"detail": "Unauthorized: Empty token"})

    def test_authentication_fail_invalid_token(self):
        headers = {"Authorization": "Bearer invalid-token"}
        post_data = {"name": "Reserve"}

        response = self.client.put(
            f"/api/admin/bank_account/{self.bank_account.uuid}/",
            data=post_data,
            headers=headers,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 401)
        self.assertJSONEqual(response.content, {"detail": "Unauthorized: Invalid token"})

    def test_check_permission_scopes_fail(self):
        # Required permission scopes for update bank account endpoint are: VIEW, EDIT_ACCOUNT
        data: list[tuple[list[ApiPermissionScope], str]] = [
            ([], "Access denied - missing permission scope(s): edit_account, view"),
            ([ApiPermissionScope.VIEW], "Access denied - missing permission scope(s): edit_account"),
            ([ApiPermissionScope.EDIT_ACCOUNT], "Access denied - missing permission scope(s): view"),
            (
                [ApiPermissionScope.VIEW, ApiPermissionScope.VIEW_IMPORT],
                "Access denied - missing permission scope(s): edit_account",
            ),
        ]
        for scopes, message in data:
            with self.subTest(scopes=scopes):
                token = create_token(TokenData(scopes=scopes, log="LOG_SESS_ID"))
                headers = {"Authorization": f"Bearer {token}"}
                post_data = {"name": "Reserve"}

                response = self.client.put(
                    f"/api/admin/bank_account/{self.bank_account.uuid}/",
                    data=post_data,
                    headers=headers,
                    content_type="application/json",
                )

                self.assertEqual(response.status_code, 403)
                self.assertJSONEqual(response.content, {"detail": message})

    def test_authentication_and_permissions_ok(self):
        token = create_token(
            TokenData(scopes=[ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_ACCOUNT], log="LOG_SESS_ID")
        )
        headers = {"Authorization": f"Bearer {token}"}
        post_data = {"name": "Reserve"}

        response = self.client.put(
            f"/api/admin/bank_account/{self.bank_account.uuid}/",
            data=post_data,
            headers=headers,
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 200)


class ChoicesPaymentProcessor(DummyPaymentProcessor):
    """Dummy processor with hints."""

    def get_client_choices(self) -> dict[str, str]:
        """Provided hints."""
        return {"hint1": "Description 1", "hint2": "Description 2"}


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestPaymentProcessor(TestCase):
    """Tests of payment processor endpoints."""

    # --- list processors ---

    @override_settings(
        PAIN_PROCESSORS={
            "success": "django_pain.tests.admin.test_forms.SuccessPaymentProcessor",
            "to_refund": "django_pain.tests.admin.test_forms.ToRefundPaymentProcessor",
            "failure": "django_pain.tests.admin.test_forms.FailurePaymentProcessor",
        }
    )
    def test_list_payment_processors(self):
        response = self.client.get("/api/admin/payment_processors/")

        self.assertEqual(response.status_code, 200)
        expected = [
            {"name": "success", "objective": "Generous bribe", "manual_tax_date": True},
            {"name": "to_refund", "objective": "Bribe to be refunded later", "manual_tax_date": False},
            {"name": "failure", "objective": "Not so generous bribe", "manual_tax_date": False},
        ]
        self.assertJSONEqual(response.content, expected)

    # --- get hints ---

    @override_settings(PAIN_PROCESSORS={})
    def test_payment_processor_hints_not_found(self):
        response = self.client.get("/api/admin/payment_processor/unknown/hints/")

        self.assertEqual(response.status_code, 404)
        self.assertJSONEqual(response.content, {"details": "Processor not found"})

    @override_settings(PAIN_PROCESSORS={"success": "django_pain.tests.admin.test_forms.SuccessPaymentProcessor"})
    def test_payment_processor_hints_none(self):
        response = self.client.get("/api/admin/payment_processor/success/hints/")

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {"choices": None})

    @override_settings(PAIN_PROCESSORS={"choices": "django_pain.tests.api.test_admin.ChoicesPaymentProcessor"})
    def test_payment_processor_hints(self):
        response = self.client.get("/api/admin/payment_processor/choices/hints/")

        self.assertEqual(response.status_code, 200)
        expected = {
            "choices": {
                "hint1": "Description 1",
                "hint2": "Description 2",
            },
        }
        self.assertJSONEqual(response.content, expected)


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestApplication(TestCase):
    """Tests of application list."""

    def test_list_applications(self):
        response = self.client.get("/api/admin/applications/")

        self.assertEqual(response.status_code, 200)
        expected = [
            {"app_id": "auctions", "objective": "Auctions of domain registration right"},
        ]
        self.assertJSONEqual(response.content, expected)
