#
# Copyright (C) 2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test ordering support for REST API lists."""

from django.test import SimpleTestCase
from ninja import Field, Schema
from pydantic import ValidationError

from django_pain.api.order import ListOrderByBase


class ExampleModel(Schema):
    """Example model for tests."""

    field_1: str
    field_2: int = Field(json_schema_extra={"order_by": "field_2"})  # type: ignore[call-arg]
    field_3: bool = Field(json_schema_extra={"order_by": "alias_3"})  # type: ignore[call-arg]


class ExampleListOrderBy(ListOrderByBase):
    """Example order definition class."""

    _model = ExampleModel
    _default = "-field_2"


class TestBankAccountRequest(SimpleTestCase):
    """Tests of bank account schema."""

    def test_validation_failure(self):
        data = [
            ("bad", "Field 'bad' is not prefixed with ordering direction sign"),
            ("+unknown", "Field 'unknown' is not included in model 'ExampleModel'"),
            ("-field_1", "Field 'field_1' is not supported for ordering"),
        ]
        for input, message in data:
            with self.subTest(input=input):
                with self.assertRaises(ValidationError) as context:
                    ExampleListOrderBy(order_by=input)  # type: ignore[call-arg]

                self.assertIn(message, repr(context.exception))

    def test_order_by_fields(self):
        data = [
            ("", ("-field_2", "id")),  # default value
            ("+field_2", ("field_2", "id")),
            ("-field_2", ("-field_2", "id")),
            ("+field_3", ("alias_3", "id")),
            ("-field_3", ("-alias_3", "id")),
            ("+field_2,-field_3", ("field_2", "-alias_3", "id")),
        ]
        for input, args in data:
            with self.subTest(input=input):
                order_by = ExampleListOrderBy(order_by=input)  # type: ignore[call-arg]
                self.assertEqual(tuple(order_by.get_arguments()), args)
