#
# Copyright (C) 2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test pagination support for REST API lists."""

from django.test import TestCase

from django_pain.api.paginate import LimitOffsetPagination
from django_pain.models import BankAccount


class TestPagination(TestCase):
    """Tests limit+offset pagination."""

    # Note: the tests use `BankAccount` model as an example

    def setUp(self):
        # Testing data for pagination and ordering
        data = [
            ("1", "B", "EUR"),
            ("2", "A", "CZK"),
            ("3", "E", "CZK"),
            ("4", "C", "CZK"),
            ("5", "D", "EUR"),
        ]
        for number, name, currency in data:
            BankAccount.objects.create(account_number=number, account_name=name, currency=currency)
        # paginator instance + queryset
        self.paginator = LimitOffsetPagination()
        self.queryset = BankAccount.objects.order_by("id").all()

    def test_pagination_with_defaults(self):
        # all items with none skipped
        input = LimitOffsetPagination.Input()  # type: ignore[call-arg]

        result = self.paginator.paginate_queryset(self.queryset, input)

        expected = [
            ("1", "B", "EUR"),
            ("2", "A", "CZK"),
            ("3", "E", "CZK"),
            ("4", "C", "CZK"),
            ("5", "D", "EUR"),
        ]
        self.assertQuerySetEqual(
            result["items"].values_list("account_number", "account_name", "currency"),
            expected,
            transform=tuple,
            ordered=False,
        )
        self.assertEqual(result["count"], 5)

    def test_pagination_with_offset(self):
        input = LimitOffsetPagination.Input(offset=3)  # type: ignore[call-arg]

        result = self.paginator.paginate_queryset(self.queryset, input)

        expected = [
            ("4", "C", "CZK"),
            ("5", "D", "EUR"),
        ]
        self.assertQuerySetEqual(
            result["items"].values_list("account_number", "account_name", "currency"),
            expected,
            transform=tuple,
            ordered=False,
        )
        self.assertEqual(result["count"], 5)

    def test_pagination_with_limit(self):
        input = LimitOffsetPagination.Input(limit=2)  # type: ignore[call-arg]

        result = self.paginator.paginate_queryset(self.queryset, input)

        expected = [
            ("1", "B", "EUR"),
            ("2", "A", "CZK"),
        ]
        self.assertQuerySetEqual(
            result["items"].values_list("account_number", "account_name", "currency"),
            expected,
            transform=tuple,
            ordered=False,
        )
        self.assertEqual(result["count"], 5)

    def test_pagination_with_limit_and_offset(self):
        input = LimitOffsetPagination.Input(offset=1, limit=2)

        result = self.paginator.paginate_queryset(self.queryset, input)

        expected = [
            ("2", "A", "CZK"),
            ("3", "E", "CZK"),
        ]
        self.assertQuerySetEqual(
            result["items"].values_list("account_number", "account_name", "currency"),
            expected,
            transform=tuple,
            ordered=False,
        )
        self.assertEqual(result["count"], 5)

    def test_pagination_with_big_limit(self):
        input = LimitOffsetPagination.Input(offset=4, limit=20)

        result = self.paginator.paginate_queryset(self.queryset, input)

        expected = [
            ("5", "D", "EUR"),
        ]
        self.assertQuerySetEqual(
            result["items"].values_list("account_number", "account_name", "currency"),
            expected,
            transform=tuple,
            ordered=False,
        )
        self.assertEqual(result["count"], 5)

    def test_pagination_with_big_offset(self):
        input = LimitOffsetPagination.Input(offset=20, limit=4)

        result = self.paginator.paginate_queryset(self.queryset, input)

        self.assertQuerySetEqual(
            result["items"].values_list("account_number", "account_name", "currency"),
            [],
            transform=tuple,
            ordered=False,
        )
        self.assertEqual(result["count"], 5)
