#
# Copyright (C) 2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test admin REST API schemas."""

from django.test import SimpleTestCase
from pydantic import ValidationError

from django_pain.api.schema import AdminBankAccountCreateRequest, AdminItemTypeCreateRequest


class TestBankAccountRequest(SimpleTestCase):
    """Tests of bank account schema."""

    def test_currency_ok(self):
        bank_account = AdminBankAccountCreateRequest(number="1234546", currency="CZK")

        self.assertEqual(bank_account.currency, "CZK")

    def test_currency_invalid(self):
        with self.assertRaises(ValidationError) as context:
            AdminBankAccountCreateRequest(number="1234546", currency="---")

        self.assertIn("Currency '---' is not supported", repr(context.exception))


class TestItemTypeRequest(SimpleTestCase):
    """Tests of billing item type schema."""

    def test_handle_ok(self):
        data = ["a", "abcd", "abcd_ef", "abcd_ef_ghi"]
        for handle in data:
            with self.subTest(handle=handle):
                AdminItemTypeCreateRequest(handle=handle, not_invoiceable=False, cpa_code="00.12")

    def test_handle_error(self):
        data = ["", "_", "_a", "a_", "a__b", "a_b_", "A_B"]
        for handle in data:
            with self.subTest(handle=handle):
                with self.assertRaises(ValidationError) as context:
                    AdminItemTypeCreateRequest(handle=handle, not_invoiceable=False, cpa_code="00.12")

                self.assertIn("String should match pattern", repr(context.exception))

    def test_cpa_code_ok(self):
        data = ["01", "01.2", "01.23", "01.23.4", "01.23.45"]
        for cpa_code in data:
            with self.subTest(cpa_code=cpa_code):
                AdminItemTypeCreateRequest(handle="example", not_invoiceable=False, cpa_code=cpa_code)

    def test_cpa_code_error(self):
        data = ["0", "012", "01.", "01.234", "01.23.45.6", "A0"]
        for cpa_code in data:
            with self.subTest(cpa_code=cpa_code):
                with self.assertRaises(ValidationError) as context:
                    AdminItemTypeCreateRequest(handle="example", not_invoiceable=False, cpa_code=cpa_code)

                self.assertIn("String should match pattern", repr(context.exception))

    def test_empty_cpa_code_ok(self):
        AdminItemTypeCreateRequest(handle="example", not_invoiceable=True, cpa_code="")

    def test_empty_cpa_code_error(self):
        with self.assertRaises(ValidationError) as context:
            AdminItemTypeCreateRequest(handle="example", not_invoiceable=False, cpa_code="")

        self.assertIn("CPA code is required for invoiceable item", repr(context.exception))
