#
# Copyright (C) 2024-2025  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test token REST API endpoints."""

from typing import Any

from django.test import SimpleTestCase, override_settings
from freezegun import freeze_time

from django_pain.api.schema import TokenData
from django_pain.api.token import TokenAuthenticationError, authenticate_token, create_token
from django_pain.constants import ApiPermissionScope
from django_pain.settings import get_admin_jwt_key


@override_settings(ROOT_URLCONF="django_pain.tests.urls")
class TestToken(SimpleTestCase):
    """Tests of access token."""

    def setUp(self):
        get_admin_jwt_key.cache_clear()

    @override_settings(PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}})
    def test_get_token_ok(self):
        data = {"scopes": ["view", "edit_account"], "log": "LOG-SESS-ID"}

        response = self.client.post("/api/token/", data=data, content_type="application/json")

        self.assertEqual(response.status_code, 200)
        jwt = authenticate_token(response.json()["token"])
        self.assertEqual(
            jwt, TokenData(scopes=[ApiPermissionScope.VIEW, ApiPermissionScope.EDIT_ACCOUNT], log="LOG-SESS-ID")
        )

    @override_settings(
        PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}, "TOKEN": "token-0123456789"}
    )
    def test_get_token_not_authenticated(self):
        data: dict[str, Any] = {"scopes": [], "log": "LOG-SESS-ID"}

        response = self.client.post("/api/token/", data=data, content_type="application/json")

        self.assertEqual(response.status_code, 401)

    @override_settings(
        PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}, "TOKEN": "token-0123456789"}
    )
    def test_get_token_with_authentication(self):
        headers = {"Authorization": "Bearer token-0123456789"}
        data: dict[str, Any] = {"scopes": ["view"], "log": "LOG-SESS-ID"}

        response = self.client.post("/api/token/", data=data, headers=headers, content_type="application/json")

        self.assertEqual(response.status_code, 200)
        jwt = authenticate_token(response.json()["token"])
        self.assertEqual(jwt, TokenData(scopes=[ApiPermissionScope.VIEW], log="LOG-SESS-ID"))

    @override_settings(PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}})
    def test_authenticate_token_expired(self):
        with freeze_time("2024-10-01T00:00Z"):
            token = create_token(TokenData(scopes=[], log="LOG-SESS-ID"))

        with freeze_time("2025-10-01T00:00Z"):
            with self.assertRaisesMessage(TokenAuthenticationError, "Token expired"):
                authenticate_token(token)

    @override_settings(PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}})
    def test_authenticate_token_empty(self):
        with self.assertRaisesMessage(TokenAuthenticationError, "Empty token"):
            authenticate_token("")

    @override_settings(PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}})
    def test_authenticate_token_invalid(self):
        token = create_token(TokenData(scopes=[], log="LOG-SESS-ID"))

        with self.assertRaisesMessage(TokenAuthenticationError, "Invalid token"):
            authenticate_token(token[:40] + "AAAA" + token[40:])

    @override_settings(PAIN_API_AUTHENTICATION={"ADMIN": {"JWT_PASSWORD": "VerySecretPassword"}})
    def test_authenticate_token_future(self):
        with freeze_time("2025-10-01T00:00Z"):
            token = create_token(TokenData(scopes=[], log="LOG-SESS-ID"))

        with freeze_time("2024-10-01T00:00Z"):
            with self.assertRaisesMessage(TokenAuthenticationError, "Invalid token"):
                authenticate_token(token)

    def test_authenticate_token_not_configured(self):
        with self.assertRaisesMessage(TokenAuthenticationError, "Authentication not configured"):
            authenticate_token("Some token")

    def test_get_token_not_configured(self):
        data: dict[str, Any] = {"scopes": [], "log": "LOG-SESS-ID"}

        response = self.client.post("/api/token/", data=data, content_type="application/json")

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {"token": ""})
