#
# Copyright (C) 2024-2025  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test checking VAT number."""

from datetime import datetime, timezone
from io import StringIO
from pathlib import Path
from typing import Any
from unittest.mock import Mock, call, patch

import requests
import requests_mock
from django.core.management import call_command
from django.test import SimpleTestCase, TestCase, override_settings
from freezegun import freeze_time
from testfixtures import LogCapture
from zeep.exceptions import Error as SoapError

from django_pain.management.commands.check_vat_number import RegistrDphClient, ViesUnresolved, validate_eu_vat_number
from django_pain.models import Customer
from django_pain.tests.utils import get_customer


class TestRegistrDphClient(SimpleTestCase):
    """Test SOAP client for RegistrDPH."""

    @classmethod
    def setUpClass(cls):
        wsdl = Path(__file__).parent / "data" / "RegistrDPH.wsdl"
        cls.regdph_client = RegistrDphClient(wsdl, timeout=3.0)
        cls.response_type = cls.regdph_client._client.get_element(
            "{http://adis.mfcr.cz/rozhraniCRPDPH/}StatusNespolehlivyPlatceResponse"
        )
        cls.status_type = cls.regdph_client._client.get_type("{http://adis.mfcr.cz/rozhraniCRPDPH/}StatusType")
        cls.info_type = cls.regdph_client._client.get_type("{http://adis.mfcr.cz/rozhraniCRPDPH/}InformaceOPlatciType")

    def test_service_ok(self):
        response = self.response_type(
            status=self.status_type(statusCode=0, statusText="OK"),
            statusPlatceDPH=[
                self.info_type(dic="0001", nespolehlivyPlatce="NE"),
                self.info_type(dic="0002", nespolehlivyPlatce="ANO"),
                self.info_type(dic="0003", nespolehlivyPlatce="NENALEZEN"),
            ],
        )
        with patch.object(self.regdph_client, "_client") as soap_client_mock:
            soap_client_mock.service.getStatusNespolehlivyPlatce.return_value = response

            response = self.regdph_client.check_batch(["CZ0001", "CZ0002", "CZ0003"])

            self.assertEqual(response, {"CZ0001": True, "CZ0002": True, "CZ0003": False})
            soap_client_mock.service.getStatusNespolehlivyPlatce.assert_called_once_with(dic=["0001", "0002", "0003"])

    def test_service_outage(self):
        response = self.response_type(
            status=self.status_type(statusCode=3, statusText="Služba nedostupná"), statusPlatceDPH=[]
        )
        with patch.object(self.regdph_client, "_client") as soap_client_mock:
            soap_client_mock.service.getStatusNespolehlivyPlatce.return_value = response

            response = self.regdph_client.check_batch(["CZ0001"])

            self.assertEqual(response, None)
            soap_client_mock.service.getStatusNespolehlivyPlatce.assert_called_once()


class TestValidateEuVatNumber(SimpleTestCase):
    """Test `validate_eu_vat_number` function."""

    VIES_URL = "https://ec.europa.eu/taxation_customs/vies/rest-api/check-vat-number"

    def test_no_vat_number(self):
        self.assertIsNone(validate_eu_vat_number(None, timeout=3.0))

    def test_non_eu_resident(self):
        self.assertIsNone(validate_eu_vat_number("CH0001", timeout=3.0))

    @requests_mock.mock()
    def test_resolved_true(self, mock):
        mock.post(self.VIES_URL, json={"valid": True})  # VAT subject found in VIES

        self.assertTrue(validate_eu_vat_number("DE0001", timeout=3.0))

        # Check HTTP call
        self.assertEqual(mock.call_count, 1)
        request = mock.last_request
        self.assertEqual(request.headers["Content-Type"], "application/json")  # check data encoding
        expected_request_data = {
            "countryCode": "DE",
            "vatNumber": "0001",
            "requesterMemberStateCode": "CZ",
            "requesterNumber": "67985726",
        }
        self.assertEqual(request.json(), expected_request_data)  # check request data

    @requests_mock.mock()
    def test_resolved_false(self, mock):
        mock.post(self.VIES_URL, json={"valid": False})  # VAT subject not found in VIES

        self.assertFalse(validate_eu_vat_number("DE0001", timeout=3.0))

    @requests_mock.mock()
    def test_unresolved(self, mock):
        # VIES check was not resolved for different reasons
        response_data = [
            {"exc": requests.exceptions.ConnectTimeout},  # exception
            {"status_code": 500},  # bad status code
            {"text": "malformed"},  # not a JSON response
            {"json": {"actionSucceed": False}},  # `valid` field not present
        ]

        for data in response_data:
            with self.subTest(data=data):
                mock.post(self.VIES_URL, **data)

                self.assertRaises(ViesUnresolved, validate_eu_vat_number, "DE0001", timeout=3.0)


@override_settings(PAIN_REGISTR_DPH_WSDL_FILE=str(Path(__file__).parent / "data" / "RegistrDPH.wsdl"))
@patch("django_pain.management.commands.check_vat_number.REGISTR_BATCH_SIZE", 2)
class TestCheckVatNumber(TestCase):
    """Test `check_vat_number` command."""

    def setUp(self):
        self.log_handler = LogCapture("django_pain.management.commands.check_vat_number", propagate=False)
        self.out = StringIO()

    def tearDown(self):
        self.log_handler.uninstall()

    @freeze_time("2024-01-10T12:00:00Z")
    @patch("django_pain.logging.get_call_id", Mock(return_value="0123"))
    @patch("django_pain.management.commands.check_vat_number.RegistrDphClient.check_batch")
    @patch("django_pain.management.commands.check_vat_number.validate_eu_vat_number")
    def test_check(self, validate_mock, regdph_mock):
        customer_data: list[dict[str, Any]] = [
            {"uid": "cid-1", "vat_number": None},  # not a VAT subject
            {"uid": "cid-2", "vat_number": "DE0002"},  # new check
            {"uid": "cid-3", "vat_number": "DE0003"},  # new check
            {"uid": "cid-4", "country_code": "CH", "vat_number": "CH0004"},  # not an EU resident
            {
                "uid": "cid-5",
                "vat_number": "DE0005",
                "vies_checked_at": datetime(2023, 12, 1, tzinfo=timezone.utc),
            },  # periodic check
            {
                "uid": "cid-6",
                "vat_number": "DE0006",
                "vies_checked_at": datetime(2023, 12, 15, tzinfo=timezone.utc),
            },  # periodic check
            {
                "uid": "cid-7",
                "vat_number": "DE0007",
                "vies_checked_at": datetime(2024, 1, 5, tzinfo=timezone.utc),
            },  # recent enough
            {"uid": "cid-8", "vat_number": "CZ0008"},  # new check
            {"uid": "cid-9", "vat_number": "CZ0009"},  # new check
            {
                "uid": "cid-10",
                "vat_number": "CZ0010",
                "vies_checked_at": datetime(2023, 12, 15, tzinfo=timezone.utc),
            },  # periodic check
            {
                "uid": "cid-11",
                "vat_number": "CZ0011",
                "vies_checked_at": datetime(2024, 1, 5, tzinfo=timezone.utc),
            },  # recent enough
        ]
        for item in customer_data:
            customer = get_customer(**item)
            customer.save()
        regdph_mock.side_effect = [{"CZ0008": True, "CZ0009": False}, {"CZ0010": True}]
        validate_mock.side_effect = [True, False, None, ViesUnresolved("Malice in Wonderland!")]

        call_command("check_vat_number", ["--interval=10", "--timeout=3.0"], stdout=self.out)

        self.assertEqual(
            regdph_mock.mock_calls,
            [
                call(["CZ0008", "CZ0009"]),
                call(["CZ0010"]),
            ],
        )
        self.assertEqual(
            validate_mock.mock_calls,
            [
                call("DE0002", timeout=3.0),
                call("DE0003", timeout=3.0),
                call("DE0005", timeout=3.0),
                call("DE0006", timeout=3.0),
            ],
        )
        self.log_handler.check_present(
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` started"),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-8 - VIES info resolved, valid: True",
            ),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-9 - VIES info resolved, valid: False",
            ),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-10 - VIES info resolved, valid: True",
            ),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by RegistrDPH: 3"),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-2 - VIES info resolved, valid: True",
            ),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-3 - VIES info resolved, valid: False",
            ),
            ("django_pain.management.commands.check_vat_number", "DEBUG", "[0123] Customer cid-5 - VIES info cleared"),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-6 - VIES info not resolved: Malice in Wonderland!",
            ),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by VIES: 3"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers not resolved: 1"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` finished"),
        )
        expected = [
            ("cid-1", None, None),
            ("cid-2", True, datetime(2024, 1, 10, 12, 0, tzinfo=timezone.utc)),
            ("cid-3", False, datetime(2024, 1, 10, 12, 0, tzinfo=timezone.utc)),
            ("cid-4", None, None),
            ("cid-5", None, None),
            ("cid-6", None, datetime(2023, 12, 15, 0, 0, tzinfo=timezone.utc)),
            ("cid-7", None, datetime(2024, 1, 5, 0, 0, tzinfo=timezone.utc)),
            ("cid-8", True, datetime(2024, 1, 10, 12, 0, tzinfo=timezone.utc)),
            ("cid-9", False, datetime(2024, 1, 10, 12, 0, tzinfo=timezone.utc)),
            ("cid-10", True, datetime(2024, 1, 10, 12, 0, tzinfo=timezone.utc)),
            ("cid-11", None, datetime(2024, 1, 5, 0, 0, tzinfo=timezone.utc)),
        ]
        self.assertQuerySetEqual(
            Customer.objects.values_list("uid", "vies_vat_valid", "vies_checked_at"),
            expected,
            transform=tuple,
            ordered=False,
        )

    @freeze_time("2024-01-10T12:00:00Z")
    @patch("django_pain.logging.get_call_id", Mock(return_value="0123"))
    @patch("django_pain.management.commands.check_vat_number.RegistrDphClient.check_batch")
    @patch("django_pain.management.commands.check_vat_number.validate_eu_vat_number")
    def test_check_regdph_outage(self, validate_mock, regdph_mock):
        # customer is not resolved by RegistrDPH (outage), resolved by VIES
        customer = get_customer(uid="cid-1", vat_number="CZ0001")
        customer.save()
        regdph_mock.return_value = None  # not resolved in RegistrDPH
        validate_mock.return_value = True  # resolved by VIES

        call_command("check_vat_number", ["--interval=10", "--timeout=3.0"], stdout=self.out)

        self.assertEqual(regdph_mock.mock_calls, [call(["CZ0001"])])
        self.assertEqual(validate_mock.mock_calls, [call("CZ0001", timeout=3.0)])
        self.log_handler.check_present(
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` started"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by RegistrDPH: 0"),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-1 - VIES info resolved, valid: True",
            ),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by VIES: 1"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers not resolved: 0"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` finished"),
        )

    @freeze_time("2024-01-10T12:00:00Z")
    @patch("django_pain.logging.get_call_id", Mock(return_value="0123"))
    @patch("django_pain.management.commands.check_vat_number.RegistrDphClient.check_batch")
    @patch("django_pain.management.commands.check_vat_number.validate_eu_vat_number")
    def test_check_regdph_error(self, validate_mock, regdph_mock):
        # customer is not resolved by RegistrDPH (SOAP error), resolved by VIES
        customer = get_customer(uid="cid-1", vat_number="CZ0001")
        customer.save()
        regdph_mock.side_effect = SoapError("Broken!")
        validate_mock.return_value = True

        call_command("check_vat_number", ["--interval=10", "--timeout=3.0"], stdout=self.out)

        self.assertEqual(regdph_mock.mock_calls, [call(["CZ0001"])])
        self.assertEqual(validate_mock.mock_calls, [call("CZ0001", timeout=3.0)])
        self.log_handler.check_present(
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` started"),
            (
                "django_pain.management.commands.check_vat_number",
                "WARNING",
                "[0123] RegistrDPH error: Error('Broken!')",
            ),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by RegistrDPH: 0"),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-1 - VIES info resolved, valid: True",
            ),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by VIES: 1"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers not resolved: 0"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` finished"),
        )

    @freeze_time("2024-01-10T12:00:00Z")
    @patch("django_pain.logging.get_call_id", Mock(return_value="0123"))
    @patch("django_pain.management.commands.check_vat_number.RegistrDphClient.__init__")
    @patch("django_pain.management.commands.check_vat_number.validate_eu_vat_number")
    def test_check_regdph_init_fail(self, validate_mock, regdph_mock):
        # creation of SOAP client fails
        customer = get_customer(uid="cid-1", vat_number="CZ0001")
        customer.save()
        regdph_mock.side_effect = requests.HTTPError("WSDL file not found!")
        validate_mock.return_value = True

        call_command("check_vat_number", ["--interval=10", "--timeout=3.0"], stdout=self.out)

        self.assertEqual(
            regdph_mock.mock_calls, [call(wsdl=str(Path(__file__).parent / "data" / "RegistrDPH.wsdl"), timeout=3.0)]
        )
        self.assertEqual(validate_mock.mock_calls, [call("CZ0001", timeout=3.0)])
        self.log_handler.check_present(
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` started"),
            (
                "django_pain.management.commands.check_vat_number",
                "WARNING",
                "[0123] RegistrDPH error: HTTPError('WSDL file not found!')",
            ),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by RegistrDPH: 0"),
            (
                "django_pain.management.commands.check_vat_number",
                "DEBUG",
                "[0123] Customer cid-1 - VIES info resolved, valid: True",
            ),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers resolved by VIES: 1"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Customers not resolved: 0"),
            ("django_pain.management.commands.check_vat_number", "INFO", "[0123] Command `check_vat_number` finished"),
        )
