#
# Copyright (C) 2019-2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test apps."""

from unittest.mock import call, patch

from django.apps import apps
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.management import call_command
from django.test import TestCase, override_settings

from django_pain.constants import LOGGER_SERVICE, LOGGER_SERVICE_HANDLE, LogEntryType, LogReferenceType, LogResult
from django_pain.models import BankPayment
from django_pain.tests.mixins import CacheResetMixin


class TestDjangoPainConfig(CacheResetMixin, TestCase):
    """Test DjangoPainConfig."""

    @patch("django_pain.apps.LOGGER.client", autospec=True)
    def test_register_logger(self, logger_client_mock):
        apps.clear_cache()
        with override_settings(INSTALLED_APPS=["django_pain.apps.DjangoPainConfig"]):
            self.assertEqual(
                logger_client_mock.mock_calls,
                [
                    call.register_service(LOGGER_SERVICE, handle=LOGGER_SERVICE_HANDLE),
                    call.register_results(LOGGER_SERVICE, results=LogResult),
                    call.register_log_entry_types(LOGGER_SERVICE, log_entry_types=LogEntryType),
                    call.register_object_reference_types(reference_types=LogReferenceType),
                ],
            )
        apps.clear_cache()

    @override_settings(PAIN_PROCESSORS={"dummy": "django_pain.tests.utils.DummyPaymentProcessor"})
    def test_post_migrate(self):
        content_type = ContentType.objects.get_for_model(BankPayment)
        results = Permission.objects.filter(codename="can_manually_assign_to_dummy", content_type=content_type)
        self.assertEqual(results.count(), 0)

        call_command("migrate", "-v0")

        results = Permission.objects.filter(codename="can_manually_assign_to_dummy", content_type=content_type)
        self.assertEqual(results.count(), 1)
        self.assertQuerySetEqual(
            results.values_list("name"), [("Can manually assign to Dummy objective",)], transform=tuple, ordered=False
        )
